from __future__ import annotations

import http.client
import threading
import time
import traceback
import uuid
from typing import TYPE_CHECKING, Callable, Protocol
from urllib.parse import urlparse

from lxml import etree

from sdc11073 import loghelper
from sdc11073 import observableproperties as properties
from sdc11073.namespaces import EventingActions
from sdc11073.pysoap.soapclient import HTTPException, HTTPReturnCodeError
from sdc11073.pysoap.soapenvelope import SoapResponseError
from sdc11073.xml_types import eventing_types as evt_types
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

from .request_handler_deferred import EmptyResponse

if TYPE_CHECKING:
    from sdc11073.definitions_base import AbstractDataModel
    from sdc11073.dispatch import RequestData
    from sdc11073.pysoap.msgfactory import CreatedMessage, MessageFactory
    from sdc11073.pysoap.msgreader import MessageReader
    from sdc11073.pysoap.soapclient import SoapClientProtocol
    from sdc11073.xml_types.eventing_types import FilterType
    from sdc11073.xml_types.mex_types import HostedServiceType
    from sdc11073 import xml_utils


class ConsumerSubscriptionProtocol(Protocol):
    """The interface of a subscription instance."""

    def subscribe(self, expires: int | float = 3600,  # seconds
                  any_elements: list | None = None,
                  any_attributes: dict | None = None) -> None:
        """Send a Subscribe message."""

    def renew(self, expires: int | float = 3600) -> float:
        """Send a Renew message."""

    def get_status(self) -> float:
        """Send a GetStatus message."""

    @property
    def remaining_subscription_seconds(self) -> float:
        """Return the time span until when the subscription will expire."""

    def on_notification(self, request_data: RequestData) -> CreatedMessage:
        """Process incoming notification."""

    def unsubscribe(self):
        """Send Unsubscribe message."""


class ConsumerSubscription:
    """ConsumerSubscription handles a subscription to an event source.

    It stores all key data of the subscription and can renew and unsubscribe this subscription.
    """

    notification_msg = properties.ObservableProperty()
    notification_data = properties.ObservableProperty()
    IDENT_TAG = etree.QName('http.local.com', 'MyClIdentifier')
    is_subscribed = properties.ObservableProperty(False)

    def __init__(self, msg_factory: MessageFactory,  # noqa: PLR0913
                 data_model: AbstractDataModel,
                 get_soap_client_func: Callable[[str], SoapClientProtocol],
                 dpws_hosted: HostedServiceType,
                 filter_type: FilterType,
                 notification_url: str,
                 end_to_url: str,
                 log_prefix: str):
        """Construct a ConsumerSubscription."""
        self._msg_factory = msg_factory
        self._data_model = data_model
        self._get_soap_client_func = get_soap_client_func
        self._dpws_hosted = dpws_hosted
        self._filter_type = filter_type
        self._filter_text = filter_type.text
        self.notification_url = notification_url
        self.end_to_url = end_to_url
        self.is_subscribed = False
        self._is_subscribed_lock = threading.Lock()
        self.end_status = None  # if device sent a SubscriptionEnd message, this contains the status from the message
        self.end_reason = None  # if device sent a SubscriptionEnd message, this contains the reason from the message
        self.expires_at: float = 0.0
        self.requested_expires: int | float = 0
        self.granted_expires: int | float = 0

        self.notify_to_identifier: xml_utils.LxmlElement | None = None

        self.end_to_identifier: xml_utils.LxmlElement | None = None

        self._logger = loghelper.get_logger_adapter('sdc.client.subscr', log_prefix)
        self.event_counter = 0  # for display purpose, we count notifications
        # ToDo: check if there is more than one address. In that case a clever selection is needed
        self._hosted_service_address: str = self._dpws_hosted.EndpointReference[0].Address
        self._hosted_service_path = urlparse(self._hosted_service_address).path
        self._subscription_manager_path: str | None = None
        self.subscribe_response: evt_types.SubscribeResponse | None = None

    def subscribe(self, expires: int | float = 3600,
                  any_elements: list | None = None,
                  any_attributes: dict | None = None) -> None:
        """Send a subscribe request to the provider and handle the response."""
        self._logger.info('start subscription "{}"', self.short_filter_string)  # noqa: PLE1205
        self.event_counter = 0
        self.requested_expires = expires  # saved for later renewal, we will use the same interval
        subscribe_request = evt_types.Subscribe()
        # EndTo is an optional element
        if self.end_to_url is not None:
            subscribe_request.init_end_to()
            subscribe_request.EndTo.Address = self.end_to_url
            if self.end_to_identifier is not None:
                subscribe_request.EndTo.ReferenceParameters = [self.end_to_identifier]
        # DeliveryMode always is the same value
        subscribe_request.Delivery.Mode = f'{self._data_model.ns_helper.WSE.namespace}/DeliveryModes/Push'
        # NotifyTo
        subscribe_request.Delivery.NotifyTo.Address = self.notification_url
        if self.notify_to_identifier is not None:
            subscribe_request.Delivery.NotifyTo.ReferenceParameters = [self.notify_to_identifier]
        subscribe_request.Expires = expires
        subscribe_request.Filter = self._filter_type
        # make an elementtree from subscribe_request and add any_elements and any_attributes if present
        nsh = self._data_model.ns_helper
        nsmap = nsh.partial_map(nsh.WSE, nsh.MSG, nsh.PM)
        body_node = subscribe_request.as_etree_node(subscribe_request.NODETYPE, nsmap)
        if any_elements is not None:
            body_node.extend(any_elements)
        if any_attributes is not None:
            for name, value in any_attributes.items():
                body_node.set(name, value)
        inf = HeaderInformationBlock(action=EventingActions.Subscribe,
                                     addr_to=self._hosted_service_address)
        message = self._msg_factory.mk_soap_message_etree_payload(inf, body_node)
        msg = f'subscribe {self.short_filter_string}'
        try:
            soap_client = self._get_soap_client_func(self._hosted_service_address)
            message_data = soap_client.post_message_to(self._hosted_service_path, message, msg=msg)
            try:
                self.subscribe_response = evt_types.SubscribeResponse.from_node(message_data.p_msg.msg_node)
                self.is_subscribed = True
                subscription_manager_address = self.subscribe_response.SubscriptionManager.Address
                self._subscription_manager_path = urlparse(subscription_manager_address).path
                self.granted_expires = self.subscribe_response.Expires
                self.expires_at = time.time() + self.granted_expires
                self._logger.info('Subscribe was successful: expires at {}, address="{}"',  # noqa: PLE1205
                                  self.expires_at, self.subscribe_response.SubscriptionManager.Address)
            except AttributeError as ex:
                self._logger.error('Subscribe response has unexpected content: {}',  # noqa: PLE1205
                                   message_data.p_msg.raw_data)
                self.is_subscribed = False
                raise SoapResponseError(message_data.p_msg) from ex
        except HTTPReturnCodeError as ex:
            self._logger.error('could not subscribe: %r', ex)

    def renew(self, expires: int = 3600) -> float:
        """Send a Renew request to the provider and handle the response.

        :return: the remaining time of the subscription or 0.0, if the request was not successful
        """
        with self._is_subscribed_lock:
            if not self.is_subscribed:
                return 0.0
            renew = evt_types.Renew()
            renew.Expires = expires
            dev_reference_param = self.subscribe_response.SubscriptionManager.ReferenceParameters
            subscription_manager_address = self.subscribe_response.SubscriptionManager.Address
            inf = HeaderInformationBlock(action=renew.action,
                                         addr_to=subscription_manager_address,
                                         reference_parameters=dev_reference_param)
            message = self._msg_factory.mk_soap_message(inf, payload=renew)
            try:
                soap_client = self._get_soap_client_func(subscription_manager_address)
                message_data = soap_client.post_message_to(
                    self._subscription_manager_path, message, msg='renew')
                self._logger.debug('{}', message_data.p_msg.raw_data)  # noqa: PLE1205
            except HTTPReturnCodeError as ex:
                self.is_subscribed = False
                self._logger.error('could not renew: {}', ex)  # noqa: PLE1205
            except (http.client.HTTPException, ConnectionError) as ex:
                self._logger.warning('renew failed: {}', ex)  # noqa: PLE1205
                self.is_subscribed = False
            except Exception as ex:  # noqa: BLE001
                # log any other exception as error and consider subscription to be broken
                self._logger.error('Exception in renew: {}', ex)  # noqa: PLE1205
                self.is_subscribed = False
            else:
                renew_response = evt_types.RenewResponse.from_node(message_data.p_msg.msg_node)
                self.granted_expires = renew_response.Expires
                if self.granted_expires is not None:
                    self.expires_at = time.time() + self.granted_expires
                    return self.granted_expires
                self.is_subscribed = False
                self._logger.warning('renew failed: {}',  # noqa: PLE1205
                                     etree.tostring(message_data.p_msg.body_node, pretty_print=True))
            return 0.0

    def unsubscribe(self):
        """Send an unsubscribe request to the provider and handle the response."""
        with self._is_subscribed_lock:
            if not self.is_subscribed:
                return
            request = evt_types.Unsubscribe()
            dev_reference_param = self.subscribe_response.SubscriptionManager.ReferenceParameters
            subscription_manager_address = self.subscribe_response.SubscriptionManager.Address
            inf = HeaderInformationBlock(action=request.action,
                                         addr_to=subscription_manager_address,
                                         reference_parameters=dev_reference_param)
            message = self._msg_factory.mk_soap_message(inf, payload=request)
            soap_client = self._get_soap_client_func(subscription_manager_address)
            received_message_data = soap_client.post_message_to(self._subscription_manager_path,
                                                                message, msg='unsubscribe')
            response_action = received_message_data.action
            # check response: response does not contain explicit status. If action== UnsubscribeResponse all is fine.
            if response_action == EventingActions.UnsubscribeResponse:
                self._logger.info(  # noqa: PLE1205
                    'unsubscribe: end of subscription {} was confirmed.', self.notification_url)
                self.is_subscribed = False
            else:
                self._logger.error(  # noqa: PLE1205
                    'unsubscribe: unexpected response action: {}', received_message_data.p_msg.raw_data)
                raise ValueError(f'unsubscribe: unexpected response action: {received_message_data.p_msg.raw_data}')

    def get_status(self) -> float:
        """Send a GetStatus Request to the device.

        :return: the remaining time of the subscription or 0, if the request was not successful
        """
        with self._is_subscribed_lock:
            if not self.is_subscribed:
                return 0.0
            request = evt_types.GetStatus()
            dev_reference_param = self.subscribe_response.SubscriptionManager.ReferenceParameters
            subscription_manager_address = self.subscribe_response.SubscriptionManager.Address
            inf = HeaderInformationBlock(action=request.action,
                                         addr_to=subscription_manager_address,
                                         reference_parameters=dev_reference_param)
            message = self._msg_factory.mk_soap_message(inf, payload=request)
            try:
                soap_client = self._get_soap_client_func(subscription_manager_address)
                message_data = soap_client.post_message_to(
                    self._subscription_manager_path,
                    message, msg='get_status')
            except HTTPReturnCodeError as ex:
                self.is_subscribed = False
                self._logger.error('could not get status: {}', ex)  # noqa: PLE1205
            except (http.client.HTTPException, ConnectionError) as ex:
                self.is_subscribed = False
                self._logger.warning(  # noqa: PLE1205
                    'get_status: Connection Error {} for subscription {}', ex, self._filter_text)
            except Exception as ex:  # noqa: BLE001
                # log any other exception as error and consider subscription to be broken
                self._logger.error('Exception in get_status: {}', ex)  # noqa: PLE1205
                self.is_subscribed = False
            else:
                get_status_response = evt_types.GetStatusResponse.from_node(message_data.p_msg.msg_node)
                expire_seconds = get_status_response.Expires
                if expire_seconds is None:
                    self._logger.warning(  # noqa: PLE1205
                        'get_status for {}: Could not find "Expires" node! get_status={} ', self._filter_text,
                        message_data.p_msg.raw_data)
                    raise SoapResponseError(message_data.p_msg)
                self._logger.debug('get_status for {}: Expires in {} seconds, counter = {}',  # noqa: PLE1205
                                   self._filter_text, expire_seconds, self.event_counter)
                return expire_seconds
            return 0.0

    @property
    def remaining_subscription_seconds(self) -> float:
        """Return remaining time."""
        return self.expires_at - time.time()

    def on_notification(self, request_data: RequestData) -> CreatedMessage:
        """Handle an incoming notification."""
        self.event_counter += 1
        self._logger.debug('received notification {} version= {}',  # noqa: PLE1205
                           request_data.message_data.action, request_data.message_data.mdib_version_group)
        self.notification_data = request_data.message_data
        self.notification_msg = request_data.message_data.p_msg
        return EmptyResponse()

    @property
    def short_filter_string(self) -> str:
        """Return a shorter version of the filter list with only the last elements of the actions."""
        if self._filter_text is not None and len(self._filter_text) > 0:
            return ', '.join(e.split('/')[-1] for e in self._filter_type.text.split())
        return '<unknown>'

    def __str__(self) -> str:
        if self._filter_text is not None:
            return (f'Subscription of "{self.short_filter_string}", is_subscribed={self.is_subscribed}, '
                    f'remaining time = {int(self.remaining_subscription_seconds)} sec., count={self.event_counter}')
        return ', '.join(str(a) for a in self._filter_type.any)


class ConsumerSubscriptionManagerProtocol(Protocol):
    """Factory for Subscription objects."""

    def __init__(self,  # noqa: PLR0913
                 msg_reader: MessageReader,
                 msg_factory: MessageFactory,
                 data_model: AbstractDataModel,
                 get_soap_client_func: Callable[[str], SoapClientProtocol],
                 notification_url: str,
                 end_to_url: str | None = None,
                 fixed_renew_interval: int | None = None,
                 log_prefix: str = ''):
        """Construct a ConsumerSubscriptionManager.

        :param msg_reader: xml -> internal representation
        :param msg_factory: internal representation -> xml
        :param notification_url: send to this address
        :param end_to_url: send to this address
        :param fixed_renew_interval: if set, renew is sent in this interval.
                                     if None, renew is sent when remaining time <= 50% of granted time
        :param log_prefix:
        """

    def start(self):
        """Start the subscription manager (typically starts a thread)."""

    def stop(self):
        """Stop the subscription manager."""

    def mk_subscription(self, dpws_hosted: HostedServiceType, filter_type: FilterType) -> ConsumerSubscriptionProtocol:
        """Create a subscription instance."""

    def on_subscription_end(self, request_data: RequestData) -> ConsumerSubscription | None:
        """Handle SubscriptionEnd message from provider."""

    def unsubscribe_all(self) -> bool:
        """Send Unsubscribe messages for all subscriptions."""


class ConsumerSubscriptionManager(threading.Thread,
                                  ConsumerSubscriptionManagerProtocol):  # derive from protocol to help typing.
    """Factory for Subscription objects.

    The thread periodically renews subscriptions. This tells the provider to keep the connection alive.
    Background info: providers can close a socket after a certain time without traffic.
    Periodic requests avoid the timeout

    This Implementation uses unique paths for notifications and SubscriptionEnd messages for each subscription.
    When the provider sends one of these messages, the unique url is enough to identify the corresponding
    subscription instance of the consumer. Reference parameters are not used.
    """

    def __init__(self, msg_reader: MessageReader,  # noqa: PLR0913
                 msg_factory: MessageFactory,
                 data_model: AbstractDataModel,
                 get_soap_client_func: Callable[[str], SoapClientProtocol],
                 notification_url: str,
                 end_to_url: str | None = None,
                 fixed_renew_interval: int | None = None,
                 log_prefix: str = ''):
        """Construct a ConsumerSubscriptionManager.

        :param msg_factory:
        :param notification_url:
        :param end_to_url:
        :param fixed_renew_interval: if set, renew is sent in this interval.
                                     if None, renew is sent when remaining time <= 50% of granted time
        :param log_prefix:
        """
        super().__init__(name=f'SubscriptionClient{log_prefix}')
        self.daemon = True
        self._msg_reader = msg_reader
        self._msg_factory = msg_factory
        self._data_model = data_model
        self._get_soap_client_func = get_soap_client_func
        self._renew_interval = fixed_renew_interval
        self.subscriptions: dict[str, ConsumerSubscription] = {}
        self._subscriptions_lock = threading.Lock()

        self._run = False
        self._notification_url = notification_url
        self._end_to_url = end_to_url or notification_url
        self._logger = loghelper.get_logger_adapter('sdc.client.subscrMgr', log_prefix)
        self.log_prefix = log_prefix
        self._counter = 1  # used to generate unique path for each subscription

    def stop(self):
        """Stop the thread."""
        self._run = False
        self.join(timeout=2)
        with self._subscriptions_lock:
            self.subscriptions.clear()

    def run(self):
        """Perform thread."""
        self._run = True
        try:
            if self._renew_interval is not None:
                self._fixed_renew_interval_loop()
            else:
                self._flexible_renew_interval_loop()
        finally:
            self._logger.info('terminating subscriptions check loop! self._run={}', self._run)  # noqa: PLE1205

    def _fixed_renew_interval_loop(self):
        """Renew subscriptions in a fixed period."""
        while self._run:
            try:
                for _ in range(self._renew_interval):
                    time.sleep(1)
                    if not self._run:
                        return
                with self._subscriptions_lock:
                    # copy list of subscriptions in order to release lock early
                    subscriptions = list(self.subscriptions.values())
                for subscription in subscriptions:
                    subscription.renew()
                for subscription in subscriptions:
                    self._logger.debug('{}', subscription)  # noqa: PLE1205
            except Exception:  # noqa: BLE001
                # catch all in order to keep thread running.
                self._logger.error('##### check loop: {}', traceback.format_exc())  # noqa: PLE1205

    def _flexible_renew_interval_loop(self):
        """Renew subscriptions when remaining time <= 50% of granted time."""
        while self._run:
            try:
                time.sleep(1)
                if not self._run:
                    return
                with self._subscriptions_lock:
                    # copy list of subscriptions in order to release lock early
                    subscriptions = list(self.subscriptions.values())
                for subscription in subscriptions:
                    # renew if remaining time is 50% of granted time or less.
                    if subscription.remaining_subscription_seconds <= subscription.granted_expires / 2:
                        subscription.renew()
                for subscription in subscriptions:
                    self._logger.debug('{}', subscription)  # noqa: PLE1205
            except Exception:  # noqa: BLE001
                # catch all in order to keep thread running.
                self._logger.error('##### check loop: {}', traceback.format_exc())  # noqa: PLE1205

    def mk_subscription(self, dpws_hosted: HostedServiceType, filter_type: FilterType) -> ConsumerSubscription:
        """Create a subscription instance."""
        sep = '' if self._notification_url.endswith('/') else '/'
        notification_url = f'{self._notification_url}{sep}subscr{self._counter}'
        sep = '' if self._end_to_url.endswith('/') else '/'
        end_to_url = f'{self._end_to_url}{sep}subscr{self._counter}_e'
        self._counter += 1
        subscription = ConsumerSubscription(self._msg_factory, self._data_model,
                                            self._get_soap_client_func,
                                            dpws_hosted, filter_type, notification_url,
                                            end_to_url, self.log_prefix)
        filter_ = filter_type.text
        with self._subscriptions_lock:
            self.subscriptions[filter_] = subscription
        return subscription

    def _find_subscription(self, request_data: RequestData,
                           log_prefix: str) -> ConsumerSubscription | None:
        for subscription in self.subscriptions.values():
            if subscription.end_to_url.endswith(request_data.current_path_element):
                return subscription
        self._logger.warning(  # noqa: PLE1205
            '{}: have no subscription for identifier = {}', log_prefix, request_data.current_path_element)
        return None

    def on_subscription_end(self, request_data: RequestData) -> ConsumerSubscription | None:
        """Handle SubscriptionEnd message from provider."""
        subscription_end = evt_types.SubscriptionEnd.from_node(request_data.message_data.p_msg.msg_node)
        info = f' status={subscription_end.Status} ' if subscription_end.Status else ''
        if len(subscription_end.Reason) > 0:
            if len(subscription_end.Reason) == 1:
                info += f' reason = {subscription_end.Reason[0]}'
            else:
                info += f' reasons = {subscription_end.Reason}'
        subscription = self._find_subscription(request_data,
                                               'on_subscription_end')
        if subscription is not None:
            self._logger.info('on_subscription_end: received Subscription End for {} {}',  # noqa: PLE1205
                              subscription.short_filter_string,
                              info)
            subscription.is_subscribed = False
            subscription.end_status = subscription_end.Status
            if len(subscription_end.Reason) > 0:
                subscription.end_reason = subscription_end.Reason[0]
            return subscription
        return None

    def unsubscribe_all(self) -> bool:
        """Send Unsubscribe messages for all subscriptions."""
        ret = True
        with self._subscriptions_lock:
            current_subscriptions = list(self.subscriptions.values())  # make a copy
            self.subscriptions.clear()
            for subscription in current_subscriptions:
                try:
                    subscription.unsubscribe()
                except HTTPException as ex:
                    self._logger.info('unsubscribe failed got HTTPException: {}', ex)  # noqa: PLE1205
                except Exception:  # noqa: BLE001
                    self._logger.error(  # noqa: PLE1205
                        'unsubscribe error: {}\n call stack:{} ', traceback.format_exc(), traceback.format_stack())
                    ret = False
        return ret


class ClientSubscriptionManagerReferenceParams(ConsumerSubscriptionManager):
    """Factory for Subscription objects. It uses reference parameters for identification of a subscription."""

    def mk_subscription(self, dpws_hosted: HostedServiceType, filter_type: FilterType) -> ConsumerSubscription:
        """Create a subscription instance."""
        subscription = ConsumerSubscription(self._msg_factory,
                                            self._data_model,
                                            self._get_soap_client_func,
                                            dpws_hosted, filter_type,
                                            self._notification_url,
                                            self._end_to_url, self.log_prefix)
        subscription.notify_to_identifier = etree.Element(ConsumerSubscription.IDENT_TAG)
        subscription.notify_to_identifier.text = uuid.uuid4().urn
        subscription.end_to_identifier = etree.Element(ConsumerSubscription.IDENT_TAG)
        subscription.end_to_identifier.text = uuid.uuid4().urn

        filter_ = filter_type.text
        with self._subscriptions_lock:
            self.subscriptions[filter_] = subscription
        return subscription

    def _find_subscription(self, request_data: RequestData,
                           log_prefix: str) -> ConsumerSubscription | None:
        subscr_ident_list = request_data.message_data.p_msg.header_info_block.reference_parameters
        if not subscr_ident_list:
            return None
        subscr_ident = subscr_ident_list[0]
        for subscription in self.subscriptions.values():
            if subscr_ident.text == subscription.end_to_identifier.text:
                return subscription
        self._logger.warning(  # noqa: PLE1205
            '{}}: have no subscription for identifier = {}', log_prefix, subscr_ident.text)
        return None



"""The module implements transactions for EntityProviderMdib."""
from __future__ import annotations

import copy
import time
from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, Union

from sdc11073.exceptions import ApiUsageError
from sdc11073.mdib.transactionsprotocol import (
    AnyTransactionManagerProtocol,
    TransactionItem,
    TransactionResultProtocol,
    TransactionType,
)

if TYPE_CHECKING:
    from sdc11073.loghelper import LoggerAdapter
    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorProtocol
    from sdc11073.mdib.statecontainers import AbstractMultiStateProtocol, AbstractStateProtocol

    from .entities import (
        ProviderEntity,
        ProviderInternalEntity,
        ProviderInternalMultiStateEntity,
        ProviderMultiStateEntity,
    )
    from .entity_providermdib import EntityProviderMdib, ProviderEntityType, ProviderInternalEntityType

    AnyProviderEntity = Union[ProviderEntity, ProviderMultiStateEntity,
                              ProviderInternalEntity, ProviderInternalMultiStateEntity]


class _Modification(Enum):
    insert = auto()
    update = auto()
    delete = auto()


@dataclass(frozen=True)
class DescriptorTransactionItem:
    """Transaction Item with old and new container."""

    entity: AnyProviderEntity
    modification: _Modification


def _update_multi_states(mdib: EntityProviderMdib, # noqa: C901
                         new: ProviderMultiStateEntity,
                         old: ProviderMultiStateEntity,
                         modified_handles: list[str] | None = None,
                         adjust_state_version: bool = True):
    if not (new.is_multi_state and old.is_multi_state):  # pragma: no cover
        msg = '_update_multi_states only handles context states!'
        raise ApiUsageError(msg)
    if new.handle != old.handle:  # pragma: no cover
        msg = f'_update_multi_states found different handles! new={new.handle}, old = {old.handle}'
        raise ApiUsageError(msg)
    if not modified_handles:
        modified_handles = new.states.keys()
    for handle in modified_handles:
        state_container = new.states.get(handle)
        if state_container is None:
            # a deleted state : this cannot be communicated via notification.
            # delete it internal_entity anf that is all
            if handle in old.states:
                old.states.pop(handle)
            else:  # pragma: no cover
                msg = f'invalid handle {handle}!'
                raise KeyError(msg)
            continue

        old_state = old.states.get(state_container.Handle)
        tmp = copy.deepcopy(state_container)

        if old_state is None:
            # this is a new state
            tmp.descriptor_container = old.descriptor
            tmp.DescriptorVersion = old.descriptor.DescriptorVersion
            if adjust_state_version:
                old_state_version = mdib.state_handle_version_lookup.get(tmp.Handle)
                if old_state_version:
                    tmp.StateVersion = old_state_version + 1
        elif adjust_state_version:
            tmp.StateVersion = old_state.StateVersion + 1


def _adjust_version_counters(new_entity: ProviderEntityType,
                             old_entity: ProviderInternalEntityType,
                             increment_descriptor_version: bool = False):
    if increment_descriptor_version:
        new_entity.descriptor.DescriptorVersion = old_entity.descriptor.DescriptorVersion + 1
    if new_entity.is_multi_state:
        for new_state in new_entity.states.values():
            new_state.DescriptorVersion = new_entity.descriptor.DescriptorVersion
            old_state = old_entity.states.get(new_state.Handle)
            if old_state is not None:
                new_state.StateVersion = old_state.StateVersion + 1
    else:
        new_entity.state.DescriptorVersion = new_entity.descriptor.DescriptorVersion
        new_entity.state.StateVersion = old_entity.state.StateVersion + 1


class _TransactionBase:
    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        self._mdib = provider_mdib
        # provide the new mdib version that the commit of this transaction will create
        self.new_mdib_version = provider_mdib.mdib_version + 1
        self._logger = logger
        self.metric_state_updates: dict[str, TransactionItem] = {}
        self.alert_state_updates: dict[str, TransactionItem] = {}
        self.component_state_updates: dict[str, TransactionItem] = {}
        self.context_state_updates: dict[str, TransactionItem] = {}
        self.operational_state_updates: dict[str, TransactionItem] = {}
        self.rt_sample_state_updates: dict[str, TransactionItem] = {}
        self._error = False

    def _handle_state_updates(self, state_updates_dict: dict) -> list[TransactionItem]:
        """Update mdib table and return a list of states to be sent in notifications."""
        updates_list = []
        for transaction_item in state_updates_dict.values():
            if transaction_item.old is not None and transaction_item.new is not None:
                # update
                entity = self._mdib.internal_entities[transaction_item.old.DescriptorHandle]
                if entity.descriptor.is_context_descriptor:
                    entity.states[transaction_item.old.Handle] = transaction_item.new
                else:
                    entity.state = transaction_item.new
            elif transaction_item.new is not None:
                # insert
                entity = self._mdib.internal_entities[transaction_item.new.DescriptorHandle]
                if entity.descriptor.is_context_descriptor:
                    entity.states[transaction_item.new.Handle] = transaction_item.new
                else:
                    entity.state = transaction_item.new
            else:
                # delete
                entity = self._mdib.internal_entities[transaction_item.old.DescriptorHandle]
                if entity.descriptor.is_context_descriptor:
                    entity.states.pop(transaction_item.old.Handle)

            if transaction_item.new is not None:
                updates_list.append(transaction_item.new.mk_copy(copy_node=False))
        return updates_list

    def get_state_transaction_item(self, handle: str) -> TransactionItem | None:
        """If transaction has a state with given handle, return the transaction-item, otherwise None.

        :param handle: the Handle of a context state or the DescriptorHandle in all other cases
        """
        if not handle:  # pragma: no cover
            raise ValueError('No handle for state specified')
        for lookup in (self.metric_state_updates,
                       self.alert_state_updates,
                       self.component_state_updates,
                       self.context_state_updates,
                       self.operational_state_updates,
                       self.rt_sample_state_updates):
            if handle in lookup:
                return lookup[handle]
        return None

    @property
    def error(self) -> bool:
        return self._error


class DescriptorTransaction(_TransactionBase):
    """A Transaction that allows to insert / update / delete Descriptors and to modify states related to them."""

    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(provider_mdib, logger)
        self.descriptor_updates: dict[str, DescriptorTransactionItem] = {}
        self._new_entities: dict[str, ProviderInternalEntityType] = {}

    def transaction_entity(self,
                           descriptor_handle: str) -> AnyProviderEntity | None:
        """Return the entity in open transaction if it exists.

        The descriptor can already be part of the transaction, and e.g. in pre_commit handlers of role providers
        it can be necessary to have access to it.
        """
        if not descriptor_handle:  # pragma: no cover
            raise ValueError('No handle for descriptor specified')
        tr_container = self.descriptor_updates.get(descriptor_handle)
        if tr_container is not None:
            if tr_container.modification == _Modification.delete:
                msg = f'The descriptor {descriptor_handle} is going to be deleted'
                raise ValueError(msg)
            return tr_container.entity
        return None

    def write_entity(self, # noqa: PLR0912, C901
                     entity: ProviderEntityType,
                     adjust_version_counter: bool = True):
        """Insert or update an entity."""
        descriptor_handle = entity.descriptor.Handle
        if descriptor_handle in self.descriptor_updates:  # pragma: no cover
            msg = f'Entity {descriptor_handle} already in updated set!'
            raise ValueError(msg)
        tmp = copy.copy(entity)  # cannot deepcopy entity, that would deepcopy also whole mdib
        tmp.descriptor = copy.deepcopy(entity.descriptor) # do not touch original entity of user
        if entity.is_multi_state:
            tmp.states = copy.deepcopy(entity.states) # do not touch original entity of user
        else:
            tmp.state = copy.deepcopy(entity.state) # do not touch original entity of user
        if descriptor_handle in self._mdib.internal_entities:
            # update
            if adjust_version_counter:
                old_entity = self._mdib.internal_entities[descriptor_handle]
                _adjust_version_counters(tmp, old_entity, increment_descriptor_version=True)
            self.descriptor_updates[descriptor_handle] = DescriptorTransactionItem(tmp,
                                                                                   _Modification.update)

        elif descriptor_handle in self._mdib.new_entities:
            # create
            if adjust_version_counter:
                version = self._mdib.descr_handle_version_lookup.get(descriptor_handle)
                if version is not None:
                    tmp.descriptor.DescriptorVersion = version
                if tmp.is_multi_state:
                    for state in tmp.states.values():
                        version = self._mdib.state_handle_version_lookup.get(state.Handle)
                        if version is not None:
                            state.StateVersion = version
                else:
                    version = self._mdib.state_handle_version_lookup.get(descriptor_handle)
                    if version is not None:
                        tmp.state.StateVersion = version
            self.descriptor_updates[descriptor_handle] = DescriptorTransactionItem(tmp,
                                                                                   _Modification.insert)
        else:
            # create without having internal entity
            tmp_entity = self._mdib.entities.new_entity(entity.node_type, entity.handle, entity.parent_handle)
            # replace descriptor and state in tmp_entity with values from tmp, but keep existing version counters
            descriptor_version = tmp_entity.descriptor.DescriptorVersion
            tmp_entity.descriptor = tmp.descriptor
            tmp_entity.descriptor.DescriptorVersion = descriptor_version
            if entity.is_multi_state:
                tmp_entity.states = tmp.states
                # change state versions if they were deleted before
                for handle, state in tmp_entity.states.items():
                    if handle in self._mdib.state_handle_version_lookup:
                        state.StateVersion = self._mdib.state_handle_version_lookup[handle] + 1
            else:
                state_version = tmp_entity.state.StateVersion
                tmp_entity.state = tmp.state
                tmp_entity.state.StateVersion = state_version
                tmp_entity.state.DescriptorVersion = descriptor_version
            self.descriptor_updates[descriptor_handle] = DescriptorTransactionItem(tmp_entity,
                                                                                   _Modification.insert)

    def write_entities(self,
                       entities: list[ProviderEntityType],
                       adjust_version_counter: bool = True):
        """Write entities in order parents first."""
        written_handles = []
        ent_dict = {ent.handle: ent for ent in entities}
        while len(written_handles) < len(ent_dict):
            for handle, ent in ent_dict.items():
                write_now = not (ent.parent_handle is not None
                                 and ent.parent_handle in ent_dict
                                 and ent.parent_handle not in written_handles)
                if write_now and handle not in written_handles:
                    # it has a parent, and parent has not been written yet
                    self.write_entity(ent, adjust_version_counter)
                    written_handles.append(handle)

    def remove_entity(self, entity: ProviderEntityType):
        """Remove existing descriptor from mdib."""
        if entity.handle in self.descriptor_updates: # pragma: no cover
            msg = f'Descriptor {entity.handle} already in updated set!'
            raise ValueError(msg)

        internal_entity = self._mdib.internal_entities.get(entity.handle)
        if internal_entity:
            self.descriptor_updates[entity.handle] = DescriptorTransactionItem(internal_entity,
                                                                               _Modification.delete)

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002, PLR0915, PLR0912, C901
        """Process transaction and create a TransactionResult.

        The parameter set_determination_time is only present in order to implement the interface correctly.
        Determination time is not set, because descriptors have no modification time.
        """
        proc = TransactionResult()
        if self.descriptor_updates:
            proc.new_mdib_version = self.new_mdib_version
            # need to know all to be deleted and to be created descriptors
            to_be_deleted_handles = [tr_item.entity.descriptor.Handle for tr_item in self.descriptor_updates.values()
                                     if tr_item.modification == _Modification.delete]
            to_be_created_handles = [tr_item.entity.descriptor.Handle for tr_item in self.descriptor_updates.values()
                                     if tr_item.modification == _Modification.insert]
            to_be_updated_handles = [tr_item.entity.descriptor.Handle for tr_item in self.descriptor_updates.values()
                                     if tr_item.modification == _Modification.update]

            # Remark 1:
            # handling only updated states here: If a descriptor is created, it can be assumed that the
            # application also creates the state in a transaction.
            # The state will then be transported via that notification report.
            # Maybe this needs to be reworked, but at the time of this writing it seems fine.
            #
            # Remark 2:
            # DescriptionModificationReport also contains the states that are related to the descriptors.
            # => if there is one, update its DescriptorVersion and add it to list of states that shall be sent
            # (Assuming that context descriptors (patient, location) are never changed,
            #  additional check for states in self.context_states is not needed.
            #  If this assumption is wrong, that functionality must be added!)

            # Restrict transaction to only insert, update or delete stuff. No mixes!
            # This simplifies handling a lot!
            filled_lists = [lst for lst in (to_be_deleted_handles, to_be_created_handles, to_be_updated_handles) if lst]
            if not filled_lists:
                return proc  # nothing changed
            if len(filled_lists) > 1:  # pragma: no cover
                raise ValueError('this transaction can only handle one of insert, update, delete!')

            for tr_item in self.descriptor_updates.values():
                if tr_item.modification == _Modification.insert:
                    # this is a create operation
                    new_entity = tr_item.entity

                    self._logger.debug(  # noqa: PLE1205
                        'transaction_manager: new entity Handle={}, node type={}',
                        new_entity.handle, new_entity.descriptor.NODETYPE)

                    # move temporary new internal entity to mdib
                    internal_entity = self._mdib.new_entities[new_entity.handle]
                    self._mdib.internal_entities[new_entity.handle] = internal_entity
                    del self._mdib.new_entities[new_entity.handle]

                    self._update_internal_entity(new_entity, internal_entity)

                    proc.descr_created.append(
                        internal_entity.descriptor)  # this will cause a Description Modification Report
                    state_update_list = proc.get_state_updates_list(new_entity.descriptor)

                    if internal_entity.is_multi_state:
                        state_update_list.extend(internal_entity.states)
                    else:
                        state_update_list.append(internal_entity.state)

                    if (internal_entity.parent_handle is not None
                            and internal_entity.parent_handle not in to_be_created_handles):
                        self._increment_parent_descriptor_version(proc, internal_entity)

                elif tr_item.modification == _Modification.delete:
                    # this is a delete operation
                    handle = tr_item.entity.descriptor.Handle
                    internal_entity = self._mdib.internal_entities.get(handle)
                    if internal_entity is None:
                        self._logger.info(  # noqa: PLE1205
                            'transaction_manager: cannot remove unknown descriptor Handle={}', handle)
                        continue

                    self._logger.debug(  # noqa: PLE1205
                        'transaction_manager: rm descriptor Handle={}', handle)
                    all_entities = self._mdib.xtra.get_all_entities_in_subtree(internal_entity)
                    for entity in all_entities:

                        # save last versions
                        self._mdib.descr_handle_version_lookup[
                            entity.descriptor.Handle] = entity.descriptor.DescriptorVersion
                        if entity.is_multi_state:
                            for state in entity.states.values():
                                self._mdib.state_handle_version_lookup[state.Handle] = state.StateVersion
                        else:
                            self._mdib.state_handle_version_lookup[entity.descriptor.Handle] = entity.state.StateVersion

                        self._mdib.internal_entities.pop(entity.handle)
                    proc.descr_deleted.extend([e.descriptor for e in all_entities])
                    # increment DescriptorVersion if a child descriptor is added or deleted.
                    if internal_entity.parent_handle is not None \
                            and internal_entity.parent_handle not in to_be_deleted_handles:
                        # only update parent if it is not also deleted in this transaction
                        self._increment_parent_descriptor_version(proc, internal_entity)
                else:
                    # this is an update operation
                    # it does not change tr_item.entity!
                    updated_entity = tr_item.entity
                    internal_entity = self._mdib.internal_entities[tr_item.entity.handle]
                    self._logger.debug(  # noqa: PLE1205
                        'transaction_manager: update descriptor Handle={}, DescriptorVersion={}',
                        internal_entity.handle, updated_entity.descriptor.DescriptorVersion)
                    self._update_internal_entity(updated_entity, internal_entity)
                    proc.descr_updated.append(
                        internal_entity.descriptor)  # this will cause a Description Modification Report
                    state_update_list = proc.get_state_updates_list(internal_entity.descriptor)
                    if updated_entity.is_multi_state:
                        state_update_list.extend(internal_entity.states.values())
                    else:
                        state_update_list.append(internal_entity.state)
        return proc

    def _update_internal_entity(self, modified_entity: ProviderEntityType,
                                internal_entity: ProviderInternalEntityType):
        """Write back information into internal entity."""
        internal_entity.descriptor.update_from_other_container(modified_entity.descriptor)
        if modified_entity.is_multi_state:
            _update_multi_states(self._mdib,
                                 modified_entity,
                                 internal_entity,
                                 None)
        else:
            internal_entity.state.update_from_other_container(modified_entity.state)

    def _increment_parent_descriptor_version(self, proc: TransactionResult,
                                             entity: ProviderInternalEntityType):
        """Increment version counter of descriptor and state.

        Add both to transaction result.
        """
        parent_entity = self._mdib.internal_entities.get(entity.parent_handle)
        updates_list = proc.get_state_updates_list(parent_entity.descriptor)

        if parent_entity is not None:
            parent_entity.descriptor.increment_descriptor_version()
            # parent entity can never be a multi state
            parent_entity.state.increment_state_version()

            proc.descr_updated.append(parent_entity.descriptor.mk_copy())
            updates_list.append(parent_entity.state.mk_copy())


class StateTransactionBase(_TransactionBase):
    """Base Class for all transactions that modify states (except ContextStateTransaction)."""

    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(provider_mdib, logger)
        self._state_updates = {}  # will be set to proper value in derived classes

    def has_state(self, descriptor_handle: str) -> bool:
        """Check if transaction has a state with given handle."""
        return descriptor_handle in self._state_updates

    def write_entity(self, entity: ProviderEntity, adjust_version_counter: bool = True):
        """Update the state of the entity."""
        if entity.is_multi_state:  # pragma: no cover
            msg = f'Multi-state entity not in {self.__class__.__name__}!'
            raise ApiUsageError(msg)
        if not self._is_correct_state_type(entity.state):
            msg = f'Wrong data type in transaction! {self.__class__.__name__}, {entity.state}'
            raise ApiUsageError(msg)
        descriptor_handle = entity.state.DescriptorHandle
        old_state = self._mdib.internal_entities[descriptor_handle].state
        tmp = copy.deepcopy(entity.state)  # do not touch original entity of user
        if adjust_version_counter:
            tmp.DescriptorVersion = old_state.DescriptorVersion
            tmp.StateVersion = old_state.StateVersion + 1
        self._state_updates[descriptor_handle] = TransactionItem(old=old_state,
                                                                 new=tmp)

    def write_entities(self, entities: list[ProviderEntity], adjust_version_counter: bool = True):
        """Update the states of entities."""
        for entity in entities:
            if entity.is_multi_state: # pragma: no cover
                msg = f'Multi-state entity not in {self.__class__.__name__}!'
                raise ApiUsageError(msg)
        for entity in entities:
            # check all states before writing any of them
            if not self._is_correct_state_type(entity.state):  # pragma: no cover
                msg = f'Wrong data type in transaction! {self.__class__.__name__}, {entity.state}'
                raise ApiUsageError(msg)
        for ent in entities:
            self.write_entity(ent, adjust_version_counter)

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:  # noqa: ARG004
        return False


class AlertStateTransaction(StateTransactionBase):
    """A Transaction for alert states."""

    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(provider_mdib, logger)
        self._state_updates = self.alert_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:
        """Process transaction and create a TransactionResult."""
        if set_determination_time:
            for tr_item in self._state_updates.values():
                new_state = tr_item.new
                old_state = tr_item.old
                if new_state is None or not hasattr(new_state, 'Presence'):
                    continue
                if old_state is None:
                    if new_state.Presence:
                        new_state.DeterminationTime = time.time()
                elif new_state.is_alert_condition and new_state.Presence != old_state.Presence:
                    new_state.DeterminationTime = time.time()
        proc = TransactionResult()
        if self._state_updates:
            proc.new_mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.alert_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_alert_state


class MetricStateTransaction(StateTransactionBase):
    """A Transaction for metric states (except real time samples)."""

    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(provider_mdib, logger)
        self._state_updates = self.metric_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:
        """Process transaction and create a TransactionResult."""
        if set_determination_time:
            for tr_item in self._state_updates.values():
                if tr_item.new is not None and tr_item.new.MetricValue is not None:
                    tr_item.new.MetricValue.DeterminationTime = time.time()
        proc = TransactionResult()
        if self._state_updates:
            proc.new_mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.metric_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_metric_state and not state.is_realtime_sample_array_metric_state


class ComponentStateTransaction(StateTransactionBase):
    """A Transaction for component states."""

    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(provider_mdib, logger)
        self._state_updates = self.component_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult."""
        proc = TransactionResult()
        if self._state_updates:
            proc.new_mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.comp_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_component_state


class RtStateTransaction(StateTransactionBase):
    """A Transaction for real time sample states."""

    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(provider_mdib, logger)
        self._state_updates = self.rt_sample_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult."""
        proc = TransactionResult()
        if self._state_updates:
            proc.new_mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.rt_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_realtime_sample_array_metric_state


class OperationalStateTransaction(StateTransactionBase):
    """A Transaction for operational states."""

    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(provider_mdib, logger)
        self._state_updates = self.operational_state_updates

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult."""
        proc = TransactionResult()
        if self._state_updates:
            proc.new_mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.op_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_operational_state


class ContextStateTransaction(_TransactionBase):
    """A Transaction for context states."""

    def __init__(self,
                 provider_mdib: EntityProviderMdib,
                 logger: LoggerAdapter):
        super().__init__(provider_mdib, logger)
        self._state_updates = self.context_state_updates

    def write_entity(self, entity: ProviderMultiStateEntity,
                     modified_handles: list[str],
                     adjust_version_counter: bool = True):
        """Insert or update a context state in mdib."""
        internal_entity = self._mdib.internal_entities[entity.descriptor.Handle]

        for handle in modified_handles:
            state_container = entity.states.get(handle)
            if state_container is None:
                # a deleted state : this cannot be communicated via notification.
                # delete it internal_entity anf that is all
                if handle in internal_entity.states:
                    internal_entity.states.pop(handle)
                else:  # pragma: no cover
                    msg = f'invalid handle {handle}!'
                    raise KeyError(msg)
                continue
            if not state_container.is_context_state:  # pragma: no cover
                raise ApiUsageError('Transaction only handles context states!')

            old_state = internal_entity.states.get(state_container.Handle)
            tmp = copy.deepcopy(state_container)  # do not touch original entity of user

            if old_state is None:
                # this is a new state
                tmp.descriptor_container = internal_entity.descriptor
                if adjust_version_counter:
                    tmp.DescriptorVersion = internal_entity.descriptor.DescriptorVersion
                    # look for previously existing state with same handle
                    old_state_version = self._mdib.state_handle_version_lookup.get(tmp.Handle)
                    if old_state_version:
                        tmp.StateVersion = old_state_version + 1
            # update
            elif adjust_version_counter:
                tmp.DescriptorVersion = internal_entity.descriptor.DescriptorVersion
                tmp.StateVersion = old_state.StateVersion + 1

            self._state_updates[state_container.Handle] = TransactionItem(old=old_state, new=tmp)

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:  # noqa: ARG002
        """Process transaction and create a TransactionResult."""
        proc = TransactionResult()
        if self._state_updates:
            proc.new_mdib_version = self.new_mdib_version
            updates = self._handle_state_updates(self._state_updates)
            proc.ctxt_updates.extend(updates)
        return proc

    @staticmethod
    def _is_correct_state_type(state: AbstractStateProtocol) -> bool:
        return state.is_context_state


class TransactionResult:
    """The transaction result.

    Data is used to create notifications.
    """

    def __init__(self):
        # states and descriptors that were modified are stored here:
        self.new_mdib_version: int | None = None
        self.descr_updated: list[AbstractDescriptorProtocol] = []
        self.descr_created: list[AbstractDescriptorProtocol] = []
        self.descr_deleted: list[AbstractDescriptorProtocol] = []
        self.metric_updates: list[AbstractStateProtocol] = []
        self.alert_updates: list[AbstractStateProtocol] = []
        self.comp_updates: list[AbstractStateProtocol] = []
        self.ctxt_updates: list[AbstractMultiStateProtocol] = []
        self.op_updates: list[AbstractStateProtocol] = []
        self.rt_updates: list[AbstractStateProtocol] = []

    @property
    def has_descriptor_updates(self) -> bool:
        """Return True if at least one descriptor is in result."""
        return len(self.descr_updated) > 0 or len(self.descr_created) > 0 or len(self.descr_deleted) > 0

    def all_states(self) -> list[AbstractStateProtocol]:
        """Return all states in this transaction."""
        return self.metric_updates + self.alert_updates + self.comp_updates + self.ctxt_updates \
            + self.op_updates + self.rt_updates

    def get_state_updates_list(self, descriptor: AbstractDescriptorProtocol) -> list:
        """Return the list that stores updated states of this descriptor."""
        if descriptor.is_context_descriptor:
            return self.ctxt_updates
        if descriptor.is_alert_descriptor:
            return self.alert_updates
        if descriptor.is_realtime_sample_array_metric_descriptor:
            return self.rt_updates
        if descriptor.is_metric_descriptor:
            return self.metric_updates
        if descriptor.is_operational_descriptor:
            return self.op_updates
        if descriptor.is_component_descriptor:
            return self.comp_updates
        msg = f'do not know how to handle {descriptor}'  # pragma: no cover
        raise ValueError(msg)  # pragma: no cover


_transaction_type_lookup = {TransactionType.descriptor: DescriptorTransaction,
                            TransactionType.alert: AlertStateTransaction,
                            TransactionType.metric: MetricStateTransaction,
                            TransactionType.operational: OperationalStateTransaction,
                            TransactionType.context: ContextStateTransaction,
                            TransactionType.component: ComponentStateTransaction,
                            TransactionType.rt_sample: RtStateTransaction}


def mk_transaction(provider_mdib: EntityProviderMdib,
                   transaction_type: TransactionType,
                   logger: LoggerAdapter) -> AnyTransactionManagerProtocol:
    """Create a transaction according to transaction_type."""
    return _transaction_type_lookup[transaction_type](provider_mdib, logger)



from sdc11073.provider.porttypes.containmenttreeserviceimpl import ContainmentTreeService
from sdc11073.provider.porttypes.contextserviceimpl import ContextService
from sdc11073.provider.porttypes.descriptioneventserviceimpl import DescriptionEventService
from sdc11073.provider.porttypes.getserviceimpl import GetService
from sdc11073.provider.porttypes.localizationservice import LocalizationService
from sdc11073.provider.porttypes.setserviceimpl import SetService
from sdc11073.provider.porttypes.stateeventserviceimpl import StateEventService
from sdc11073.provider.porttypes.waveformserviceimpl import WaveformService
from sdc11073.provider.providerimpl import SdcProvider

__all__ = [
    'SdcProvider',
    'ContainmentTreeService',
    'ContextService',
    'DescriptionEventService',
    'GetService',
    'LocalizationService',
    'SetService',
    'StateEventService',
    'WaveformService',
]



"""Implementation of data types used in Participant Model."""
from __future__ import annotations

import inspect
import sys
from dataclasses import dataclass
from decimal import Decimal
from typing import TYPE_CHECKING

from sdc11073.namespaces import QN_TYPE, text_to_qname

from . import ext_qnames as ext
from . import msg_qnames as msg
from . import pm_qnames as pm
from . import xml_structure as cp
from .basetypes import StringEnum, XMLTypeBase

if TYPE_CHECKING:
    from lxml import etree

    from sdc11073 import xml_utils
    from sdc11073.xml_types.isoduration import DateTypeUnion, DurationType


class SafetyClassification(StringEnum):
    """Represents BICEPS SafetyClassification."""

    INF = 'Inf'
    MED_A = 'MedA'
    MED_B = 'MedB'
    MED_C = 'MedC'


class MdsOperatingMode(StringEnum):
    """Represents BICEPS MdsOperatingMode."""

    NORMAL = 'Nml'
    DEMO = 'Dmo'
    SERVICE = 'Srv'
    MAINTENANCE = 'Mtn'


class ComponentActivation(StringEnum):
    """Represents BICEPS ComponentActivation."""

    ON = 'On'
    NOT_READY = 'NotRdy'
    STANDBY = 'StndBy'
    OFF = 'Off'
    SHUTDOWN = 'Shtdn'
    FAILURE = 'Fail'


class ContextAssociation(StringEnum):
    """Represents BICEPS ContextAssociation."""

    NO_ASSOCIATION = 'No'
    PRE_ASSOCIATION = 'Pre'
    ASSOCIATED = 'Assoc'
    DISASSOCIATED = 'Dis'


class AlertConditionMonitoredLimits(StringEnum):
    """Represents BICEPS AlertConditionMonitoredLimits."""

    ALL_ON = 'All'
    LOW_OFF = 'LoOff'
    HIGH_OFF = 'HiOff'
    NONE = 'None'


class AlertConditionPriority(StringEnum):
    """Represents BICEPS AlertConditionPriority."""

    NONE = 'None'
    LOW = 'Lo'
    MEDIUM = 'Me'
    HIGH = 'Hi'


class AlertConditionKind(StringEnum):
    """Represents BICEPS AlertConditionKind."""

    PHYSIOLOGICAL = 'Phy'
    TECHNICAL = 'Tec'
    OTHER = 'Oth'


class CanEscalate(StringEnum):
    """Represents BICEPS CanEscalate."""

    LOW = 'Lo'
    MEDIUM = 'Me'
    HIGH = 'Hi'


class CanDeEscalate(StringEnum):
    """Represents BICEPS CanDeEscalate."""

    MEDIUM = 'Me'
    LOW = 'Lo'
    NONE = 'None'


class AlertSignalPresence(StringEnum):
    """Represents BICEPS AlertSignalPresence."""

    ON = 'On'
    OFF = 'Off'
    LATCH = 'Latch'
    ACK = 'Ack'


class MetricAvailability(StringEnum):
    """Represents BICEPS MetricAvailability."""

    INTERMITTENT = 'Intr'
    CONTINUOUS = 'Cont'


class MetricCategory(StringEnum):
    """Represents BICEPS MetricCategory."""

    UNSPECIFIED = 'Unspec'
    MEASUREMENT = 'Msrmt'
    CALCULATION = 'Clc'
    SETTING = 'Set'
    PRESETTING = 'Preset'
    RECOMMENDATION = 'Rcmm'


class DerivationMethod(StringEnum):
    """Represents BICEPS DerivationMethod."""

    AUTOMATIC = 'Auto'
    MANUAL = 'Man'


class AccessLevel(StringEnum):
    """Represents BICEPS AccessLevel."""

    USER = 'Usr'
    CLINICAL_SUPER_USER = 'CSUsr'
    RESPONSIBLE_ORGANIZATION = 'RO'
    SERVICE_PERSONNEL = 'SP'
    OTHER = 'Oth'


class AlertSignalPrimaryLocation(StringEnum):
    """Represents BICEPS AlertSignalPrimaryLocation."""

    LOCAL = 'Loc'
    REMOTE = 'Rem'


class LocalizedTextWidth(StringEnum):
    """Represents BICEPS LocalizedTextWidth."""

    XS = 'xs'
    S = 's'
    M = 'm'
    L = 'l'
    XL = 'xl'
    XXL = 'xxl'


class OperatingMode(StringEnum):
    """Represents BICEPS OperatingMode."""

    DISABLED = 'Dis'
    ENABLED = 'En'
    NA = 'NA'


OperatingModeType = OperatingMode  # TypeAlias


class MeasurementValidity(StringEnum):
    """Represents BICEPS MeasurementValidity."""

    VALID = 'Vld'
    VALIDATED_DATA = 'Vldated'
    MEASUREMENT_ONGOING = 'Ong'
    QUESTIONABLE = 'Qst'
    CALIBRATION_ONGOING = 'Calib'
    INVALID = 'Inv'
    OVERFLOW = 'Oflw'
    UNDERFLOW = 'Uflw'
    NA = 'NA'


class GenerationMode(StringEnum):
    """Represents BICEPS MeasurementValidity.

    Describes whether METRIC data is generated by real measurements or under unreal settings (demo or test data).
    """

    REAL = 'Real'  # Real Data. A value that is generated under real conditions
    TEST = 'Test'  # Test Data. A value that is arbitrary and is for testing purposes only
    DEMO = 'Demo'  # Demo Data. A value that is arbitrary and is for demonstration purposes only


class AlertSignalManifestation(StringEnum):
    """Represents BICEPS AlertSignalManifestation."""

    AUD = 'Aud'  # Aud = Audible. The ALERT SIGNAL manifests in an audible manner, i.e., the alert can be heard.
    #  Example: an alarm sound.
    VIS = 'Vis'  # Vis = Visible. The ALERT SIGNAL manifests in a visible manner, i.e., the alert can be seen.
    #  Example: a red flashing light.
    TAN = 'Tan'  # Tan = Tangible. The ALERT SIGNAL manifests in a tangible manner, i.e., the alert can be felt.
    #  Example: vibration.
    OTH = 'Oth'  # Oth = Other. The ALERT SIGNAL manifests in a manner not further specified.


class AlertActivation(StringEnum):
    """Represents BICEPS AlertActivation."""

    ON = 'On'
    OFF = 'Off'
    PAUSED = 'Psd'


class AbstractMetricDescriptorRelationKindEnum(StringEnum):
    """Represents BICEPS AbstractMetricDescriptor/Relation/Kind Definition."""

    RECOMMENDATION = 'Rcm'
    PRE_SETTING = 'PS'
    SET_OF_SUMMARY_STATISTICS = 'SST'
    EFFECT_ON_CONTAINMENT_TREE_ENTRIES = 'ECE'
    DERIVED_FROM_CONTAINMENT_TREE_ENTRIES = 'DCE'
    OTHER = 'Oth'


class PatientType(StringEnum):
    """Represents BICEPS PatientType."""

    UNSPECIFIED = 'Unspec'
    ADULT = 'Ad'
    ADOLESCENT = 'Ado'
    PEDIATRIC = 'Ped'
    INFANT = 'Inf'
    NEONATAL = 'Neo'
    OTHER = 'Oth'


PatientTypeType = PatientType  # TypeAlias


class Sex(StringEnum):
    """Represents BICEPS Sex."""

    UNSPEC = 'Unspec'
    MALE = 'M'
    FEMALE = 'F'
    UNKNOWN = 'Unkn'


SexType = Sex  # TypeAlias


class CalibrationState(StringEnum):  # only used in CalibrationInfo
    """Represents BICEPS CalibrationState."""

    NOT_CALIBRATED = 'No'
    CALIBRATION_REQUIRED = 'Req'
    RUNNING = 'Run'
    CALIBRATED = 'Cal'
    OTHER = 'Oth'


class CalibrationType(StringEnum):  # only used in CalibrationInfo
    """Represents BICEPS CalibrationType."""

    OFFSET = 'Offset'
    GAIN = 'Gain'
    TWO_POINT_CALIBRATION = 'TP'
    UNSPEC = 'Unspec'


class PropertyBasedPMType(XMLTypeBase):
    """PropertyBasedPMType classes handle the XMLSchema-instance:Type attribute.

    The value_class_from_node method returns a matching specific class.
    it looks for classes with a matching NODETYPE member.
    """

    @classmethod
    def value_class_from_node(cls, node: xml_utils.LxmlElement) -> type[PropertyBasedPMType]:
        """If node has a xsi:Type attribute, return the class that reflects that type, else cls."""
        xsi_type_str = node.get(QN_TYPE)
        if xsi_type_str is None:
            return cls
        xsi_type = text_to_qname(xsi_type_str, node.nsmap)
        return _get_pmtypes_class(xsi_type)


class LocalizedText(PropertyBasedPMType):
    """Represents BICEPS LocalizedText."""

    NODETYPE = pm.LocalizedText
    text: str = cp.NodeStringProperty()  # this is the text of the node. Here attribute is lower case!
    Ref: str | None = cp.LocalizedTextRefAttributeProperty('Ref')
    Lang: str | None = cp.StringAttributeProperty('Lang')
    Version: int | None = cp.ReferencedVersionAttributeProperty('Version')
    TextWidth = cp.EnumAttributeProperty('TextWidth', enum_cls=LocalizedTextWidth)
    _props = ('text', 'Ref', 'Lang', 'Version', 'TextWidth')

    def __init__(self, text: str,
                 lang: str | None = None,
                 ref: str | None = None,
                 version: int | None = None,
                 text_width: LocalizedTextWidth | None = None):
        super().__init__()
        self.text = text
        self.Lang = lang
        self.Ref = ref
        self.Version = version
        self.TextWidth = text_width

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> LocalizedText:
        """Construct class from a node."""
        obj = cls('')
        obj.update_from_node(node)
        return obj

    def __repr__(self) -> str:
        params = [f'"{self.text}"']
        if self.Lang is not None:
            params.append(f'Lang={self.Lang}')
        if self.Ref is not None:
            params.append(f'Ref={self.Ref}')
        if self.Version is not None:
            params.append(f'Version={self.Version}')
        if self.TextWidth is not None:
            params.append(f'TextWidth={self.TextWidth}')
        return f'{self.__class__.__name__}({", ".join(params)})'


DEFAULT_CODING_SYSTEM = 'urn:oid:1.2.840.10004.1.1.1.0.0.1'  # ISO/IEC 11073-10101


@dataclass(frozen=True)
class Coding:
    """Coding contains the three Values that identify a CodedValue.

    It is a frozen class, it can be used as Key in dictionaries.
    """

    code: str
    coding_system: str = DEFAULT_CODING_SYSTEM
    coding_system_version: str | None = None

    def __post_init__(self):
        """Previous versions allowed also an int as code. This exception makes it easier to detect wrong type."""
        if not isinstance(self.code, str):
            raise TypeError('code must be a string!')

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> Coding:
        """Construct class from a node."""
        code = node.get('Code')
        coding_system = node.get('CodingSystem', DEFAULT_CODING_SYSTEM)
        coding_system_version = node.get('CodingSystemVersion')
        return cls(code, coding_system, coding_system_version)


class Translation(PropertyBasedPMType):
    """Represents BICEPS Translation."""

    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Code: str = cp.CodeIdentifierAttributeProperty('Code', is_optional=False)
    CodingSystem: str | None = cp.StringAttributeProperty('CodingSystem', implied_py_value=DEFAULT_CODING_SYSTEM)
    CodingSystemVersion: str | None = cp.StringAttributeProperty('CodingSystemVersion')

    _props = ('ExtExtension', 'Code', 'CodingSystem', 'CodingSystemVersion')

    def __init__(self, code: str,
                 coding_system: str | None = None,
                 coding_system_version: str | None = None):
        super().__init__()
        if code is not None and not isinstance(code, str):
            raise TypeError('code must be a string!')
        self.Code = code
        self.CodingSystem = coding_system
        self.CodingSystemVersion = coding_system_version

    @property
    def coding(self) -> Coding:
        """Return Coding."""
        return Coding(self.Code, self.CodingSystem, self.CodingSystemVersion)

    def __repr__(self) -> str:
        if self.CodingSystem is None:
            return f'CodedValue({self.Code})'
        if self.CodingSystemVersion is None:
            return f'CodedValue({self.Code}, codingsystem={self.CodingSystem})'
        return (f'CodedValue({self.Code}, codingsystem={self.CodingSystem}, '
                f'codingsystemversion={self.CodingSystemVersion})')

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> Translation:
        """Construct class from a node."""
        obj = cls('')
        obj.update_from_node(node)
        return obj


TranslationType = Translation  # TypeAlias


class CodedValue(PropertyBasedPMType):
    """Represents BICEPS CodedValue."""

    NODETYPE = pm.CodedValue
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    CodingSystemName: list[LocalizedText] | None = cp.SubElementListProperty(pm.CodingSystemName,
                                                                             value_class=LocalizedText)
    ConceptDescription: list[LocalizedText] | None = cp.SubElementListProperty(pm.ConceptDescription,
                                                                               value_class=LocalizedText)
    Translation: list[TranslationType] = cp.SubElementListProperty(pm.Translation, value_class=Translation)
    Code: str = cp.CodeIdentifierAttributeProperty('Code', is_optional=False)
    CodingSystem: str | None = cp.StringAttributeProperty('CodingSystem', implied_py_value=DEFAULT_CODING_SYSTEM)
    CodingSystemVersion: str | None = cp.StringAttributeProperty('CodingSystemVersion')
    SymbolicCodeName: str | None = cp.SymbolicCodeNameAttributeProperty('SymbolicCodeName')
    _props = ('ExtExtension', 'CodingSystemName', 'ConceptDescription', 'Translation',
              'Code', 'CodingSystem', 'CodingSystemVersion', 'SymbolicCodeName')

    def __init__(self, code: str,  # noqa: PLR0913
                 coding_system: str | None = None,
                 coding_system_version: str | None = None,
                 coding_system_names: list[LocalizedText] | None = None,
                 concept_descriptions: list[LocalizedText] | None = None,
                 symbolic_code_name: str | None = None):
        super().__init__()
        if code is not None and not isinstance(code, str):
            raise TypeError('code must be a string!')
        self.Code = code
        self.CodingSystem = coding_system
        self.CodingSystemVersion = coding_system_version
        self.CodingSystemName = [] if coding_system_names is None else coding_system_names
        self.ConceptDescription = [] if concept_descriptions is None else concept_descriptions
        self.SymbolicCodeName = symbolic_code_name

    @property
    def coding(self) -> Coding:
        """Return Coding."""
        return Coding(self.Code, self.CodingSystem, self.CodingSystemVersion)

    @property
    def all_codings(self) -> list[Coding]:
        """Return Coding of self plus Codings of all Translations."""
        ret = [self.coding]
        if self.Translation is not None:
            ret.extend([t.coding for t in self.Translation])
        return ret

    def __repr__(self) -> str:
        if self.CodingSystem is None:
            return f'CodedValue({self.Code})'
        if self.CodingSystemVersion is None:
            return f'CodedValue({self.Code}, codingsystem="{self.CodingSystem}")'
        return (f'CodedValue({self.Code}, codingsystem="{self.CodingSystem}", '
                f'codingsystemversion="{self.CodingSystemVersion}")')

    def is_equivalent(self, other: Coding | CodedValue) -> bool:
        """Compare with a CodedValue or a Coding.

        BICEPS specifies equivalence as follows:
        Two CodedValue objects C1 and C2 are equivalent, if
        - C1/@Code equals C2/@Code
        - C1/@CodingSystem equals C2/@CodingSystem, both with expanded default values
        - C1/@CodingSystemVersion equals C2/@CodingSystemVersion
        - If there exists a CodedValue object T1 in C1/pm:Translation and a CodedValue object T2 in C2/pm:Translation
          such that T1 and T2 are equivalent, C1 and T2 are equivalent, or C2 and T1 are equivalent.

        NOTE 1—In case that ./@CodingSystem is not explicitly defined in CodedValue, it is replaced implicitly by
               a default identifier. The ./@CodingSystem ATTRIBUTE is then called "expanded".
        NOTE 2—As prescribed in ./@CodingSystemVersion, a version is set only if a unique version identification
               by ./@CodingSystem is not possible. Hence, there can be no implicit version mismatch.
        NOTE 3—Equivalence between CodedValue objects is not necessarily transitive.
        """
        return have_matching_codes(self, other)

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> CodedValue:
        """Construct class from a node."""
        obj = cls('')
        obj.update_from_node(node)
        return obj


def have_matching_codes(code_a: CodedValue | Coding, code_b: CodedValue | Coding) -> bool:
    """Test if there is at least one common Coding in code_a and code_b."""
    try:
        codes_a = set(code_a.all_codings)
    except AttributeError:
        codes_a = {code_a}
    try:
        codes_b = set(code_b.all_codings)
    except AttributeError:
        codes_b = {code_b}
    common_codes = codes_a.intersection(codes_b)
    return len(common_codes) > 0


class Annotation(PropertyBasedPMType):
    """Represents BICEPS AbstractMetricValue/Annotation."""

    NODETYPE = pm.Annotation
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Type: CodedValue | None = cp.SubElementProperty(pm.Type, value_class=CodedValue)
    _props = ('ExtExtension', 'Type')

    def __init__(self, coded_value: CodedValue | None = None):
        super().__init__()
        self.Type = coded_value


AnnotationType = Annotation  # TypeAlias


class OperationGroup(PropertyBasedPMType):
    """Represents BICEPS ScoState/OperationGroup."""

    NODETYPE = pm.OperationGroup
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Type: CodedValue | None = cp.SubElementProperty(pm.Type, value_class=CodedValue)
    OperatingMode: OperatingModeType | None = cp.EnumAttributeProperty('OperatingMode', enum_cls=OperatingMode)
    Operations = cp.OperationRefListAttributeProperty('Operations')
    _props = ('ExtExtension', 'Type', 'OperatingMode', 'Operations')

    def __init__(self, coded_value: CodedValue | None = None,
                 operating_mode: OperatingModeType | None = None,
                 operations: list[str] | None = None):
        super().__init__()
        self.Type = coded_value
        self.OperatingMode = operating_mode
        self.Operations = operations


class InstanceIdentifier(PropertyBasedPMType):
    """Represents BICEPS InstanceIdentifier."""

    NODETYPE = pm.InstanceIdentifier
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Type: CodedValue | None = cp.SubElementProperty(pm.Type, value_class=CodedValue, is_optional=True)
    IdentifierName: list[LocalizedText] = cp.SubElementListProperty(pm.IdentifierName, value_class=LocalizedText)
    Root: str | None = cp.AnyURIAttributeProperty('Root')
    # default_py_value='biceps.uri.unk')  # default is defined in R0135
    Extension = cp.ExtensionAttributeProperty('Extension')  # a xsd:string
    _props = ('ExtExtension', 'Type', 'IdentifierName', 'Root', 'Extension')

    def __init__(self, root: str | None = None,
                 type_coded_value: CodedValue | None = None,
                 identifier_names: list[LocalizedText] | None = None,
                 extension_string: str | None = None):
        super().__init__()
        self.Root = root
        self.Type = type_coded_value
        self.IdentifierName = [] if identifier_names is None else identifier_names
        self.Extension = extension_string
        self.node = None

    def __repr__(self) -> str:
        params = [f'root={self.Root!r}']
        if self.Type is not None:
            params.append(f'Type={self.Type}')
        if len(self.IdentifierName) > 0:
            params.append(f'IdentifierName={[str(n) for n in self.IdentifierName]}')
        return f'InstanceIdentifier({", ".join(params)})'


class OperatingJurisdiction(InstanceIdentifier):
    """Represents BICEPS OperatingJurisdiction."""

    NODETYPE = pm.OperatingJurisdiction


class Range(PropertyBasedPMType):
    """Represents BICEPS Range."""

    NODETYPE = pm.Range
    Extension = cp.ExtensionNodeProperty(ext.Extension)
    Lower: Decimal | None = cp.DecimalAttributeProperty('Lower')
    Upper: Decimal | None = cp.DecimalAttributeProperty('Upper')
    StepWidth: Decimal | None = cp.DecimalAttributeProperty('StepWidth')
    RelativeAccuracy: Decimal | None = cp.DecimalAttributeProperty('RelativeAccuracy')
    AbsoluteAccuracy: Decimal | None = cp.DecimalAttributeProperty('AbsoluteAccuracy')
    _props = ('Extension', 'Lower', 'Upper', 'StepWidth', 'RelativeAccuracy', 'AbsoluteAccuracy')

    def __init__(self,
                 lower: Decimal | None = None,
                 upper: Decimal | None = None,
                 step_width: Decimal | None = None,
                 relative_accuracy: Decimal | None = None,
                 absolute_accuracy: Decimal | None = None):
        super().__init__()
        self.Lower = lower
        self.Upper = upper
        self.StepWidth = step_width
        self.RelativeAccuracy = relative_accuracy
        self.AbsoluteAccuracy = absolute_accuracy

    def __repr__(self) -> str:
        return (f'Range (Lower={self.Lower!r}, Upper={self.Upper!r}, StepWidth={self.StepWidth!r}, '
                f'RelativeAccuracy={self.RelativeAccuracy!r}, AbsoluteAccuracy={self.AbsoluteAccuracy!r})')


RangeType = Range  # TypeAlias


class Measurement(PropertyBasedPMType):
    """Represents BICEPS Measurement."""

    NODETYPE = pm.Measurement
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    MeasurementUnit: CodedValue = cp.SubElementProperty(pm.MeasurementUnit, value_class=CodedValue)
    MeasuredValue: Decimal = cp.DecimalAttributeProperty('MeasuredValue', is_optional=False)
    _props = ('ExtExtension', 'MeasurementUnit', 'MeasuredValue')

    def __init__(self,
                 value: Decimal | None,
                 unit: CodedValue | None):
        super().__init__()
        self.MeasuredValue = value
        self.MeasurementUnit = unit

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> Measurement:
        """Construct class from a node."""
        obj = cls(None, None)
        obj.update_from_node(node)
        return obj

    def __repr__(self) -> str:
        return f'Measurement(value={self.MeasuredValue!r}, Unit={self.MeasurementUnit!r})'


class AllowedValue(PropertyBasedPMType):
    """One AllowedValue of a EnumStringMetricDescriptor."""

    NODETYPE = pm.AllowedValue
    Value: str = cp.NodeStringProperty(pm.Value)
    Type: CodedValue | None = cp.SubElementProperty(pm.Type,
                                                    value_class=CodedValue,
                                                    is_optional=True)
    Identification: CodedValue | None = cp.SubElementProperty(pm.Identification,
                                                              value_class=InstanceIdentifier,
                                                              is_optional=True)
    Characteristic: Measurement | None = cp.SubElementProperty(pm.Characteristic,
                                                               value_class=Measurement,
                                                               is_optional=True)
    _props = ('Value', 'Type', 'Identification', 'Characteristic')

    def __init__(self, value: str = '',
                 type_coding: CodedValue | None = None,
                 identification: CodedValue | None = None,
                 characteristic: Measurement | None = None):
        """Construct one AllowedValue of a EnumStringMetricDescriptor."""
        super().__init__()
        self.Value = value
        self.Type = type_coding
        self.Identification = identification
        self.Characteristic = characteristic


class MetricQuality(PropertyBasedPMType):
    """Represents BICEPS AbstractMetricValue/MetricQuality."""

    NODETYPE = pm.MetricQuality
    Validity: MeasurementValidity = cp.EnumAttributeProperty('Validity',
                                                             enum_cls=MeasurementValidity,
                                                             is_optional=False,
                                                             default_py_value=MeasurementValidity.VALID)
    Mode: GenerationMode | None = cp.EnumAttributeProperty('Mode',
                                                           implied_py_value=GenerationMode.REAL,
                                                           enum_cls=GenerationMode)
    Qi: Decimal | None = cp.QualityIndicatorAttributeProperty('Qi', implied_py_value=Decimal('1'))
    _props = ('Validity', 'Mode', 'Qi')


MetricQualityType = MetricQuality  # TypeAlias


class AbstractMetricValue(PropertyBasedPMType):
    """Represents BICEPS AbstractMetricValue."""

    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    StartTime: float = cp.TimestampAttributeProperty('StartTime')
    StopTime: float = cp.TimestampAttributeProperty('StopTime')
    DeterminationTime: float = cp.TimestampAttributeProperty('DeterminationTime')
    MetricQuality: MetricQualityType = cp.SubElementProperty(pm.MetricQuality,
                                                             value_class=MetricQuality,
                                                             default_py_value=MetricQuality())
    Annotation: list[AnnotationType] = cp.SubElementListProperty(pm.Annotation, Annotation)
    _props = ('ExtExtension', 'StartTime', 'StopTime', 'DeterminationTime', 'MetricQuality', 'Annotation')


class NumericMetricValue(AbstractMetricValue):
    """Represents BICEPS NumericMetricValue."""

    NODETYPE = pm.NumericMetricValue
    Value: Decimal | None = cp.DecimalAttributeProperty('Value')
    _props = ('Value',)

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__} Validity={self.MetricQuality.Validity}'
                f' Value={self.Value} DeterminationTime={self.DeterminationTime}')


class StringMetricValue(AbstractMetricValue):
    """Represents BICEPS StringMetricValue."""

    NODETYPE = pm.StringMetricValue
    Value: str | None = cp.StringAttributeProperty('Value')
    _props = ('Value',)

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__} Validity={self.MetricQuality.Validity} '
                f'Value={self.Value} DeterminationTime={self.DeterminationTime}')


class ApplyAnnotation(PropertyBasedPMType):
    """Represents BICEPS SampleArrayValue/ApplyAnnotation."""

    NODETYPE = pm.ApplyAnnotation
    AnnotationIndex: int = cp.UnsignedIntAttributeProperty('AnnotationIndex', is_optional=False)
    SampleIndex: int = cp.UnsignedIntAttributeProperty('SampleIndex', is_optional=False)
    _props = ('AnnotationIndex', 'SampleIndex')

    def __init__(self, annotation_index: int | None = None, sample_index: int | None = None):
        super().__init__()
        self.AnnotationIndex = annotation_index
        self.SampleIndex = sample_index

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(AnnotationIndex={self.AnnotationIndex}, SampleIndex={self.SampleIndex})'


ApplyAnnotationType = ApplyAnnotation  # TypeAlias


class SampleArrayValue(AbstractMetricValue):
    """Represents BICEPS SampleArrayValue."""

    NODETYPE = pm.SampleArrayValue
    Samples: list[Decimal] = cp.DecimalListAttributeProperty('Samples')  # list of xs:decimal types
    ApplyAnnotation: list[ApplyAnnotationType] = cp.SubElementListProperty(pm.ApplyAnnotation, ApplyAnnotation)
    _props = ('Samples', 'ApplyAnnotation')

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} Samples={self.Samples} ApplyAnnotation={self.ApplyAnnotation}'


class RemedyInfo(PropertyBasedPMType):
    """Represents BICEPS RemedyInfo."""

    NODETYPE = pm.RemedyInfo
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Description: list[LocalizedText] = cp.SubElementListProperty(pm.Description, value_class=LocalizedText)
    _props = ('ExtExtension', 'Description')

    def __init__(self, descriptions: list[LocalizedText] | None = None):
        super().__init__()
        if descriptions:
            self.Description = descriptions


RemedyInfoType = RemedyInfo  # TypeAlias


class CauseInfo(PropertyBasedPMType):
    """Represents BICEPS CauseInfo."""

    NODETYPE = pm.CauseInfo
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    RemedyInfo: RemedyInfoType | None = cp.SubElementProperty(pm.RemedyInfo, value_class=RemedyInfo)
    Description: list[LocalizedText] = cp.SubElementListProperty(pm.Description, value_class=LocalizedText)
    _props = ('ExtExtension', 'RemedyInfo', 'Description')

    def __init__(self, remedy_info: RemedyInfoType | None = None,
                 descriptions: list[LocalizedText] | None = None):
        super().__init__()
        self.RemedyInfo = remedy_info
        self.Description = descriptions or []


class ActivateOperationDescriptorArgument(PropertyBasedPMType):
    """Represents BICEPS AbstractSetStateOperationDescriptor/Argument."""

    ArgName: CodedValue = cp.SubElementProperty(pm.ArgName, value_class=CodedValue, is_optional=False)
    Arg: xml_utils.QName | None = cp.NodeTextQNameProperty(pm.Arg, is_optional=False)
    _props = ('ArgName', 'Arg')

    def __init__(self, arg_name: CodedValue | None = None, arg: xml_utils.QName | None = None):
        super().__init__()
        self.ArgName = arg_name
        self.Arg = arg

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(argName={self.ArgName}, arg={self.Arg})'


class PhysicalConnectorInfo(PropertyBasedPMType):
    """Represents BICEPS PhysicalConnectorInfo.

    PhysicalConnectorInfo defines a number in order to allow to guide the clinical user for a failure,
    e.g., in case of a disconnection of a sensor or an ultrasonic handpiece.
    """

    NODETYPE = pm.PhysicalConnectorInfo
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Label: list[LocalizedText] = cp.SubElementListProperty(pm.Label,
                                                           value_class=LocalizedText)
    Number: int | None = cp.IntegerAttributeProperty('Number')
    _props = ('ExtExtension', 'Label', 'Number')

    def __init__(self, labels: list[LocalizedText] | None = None,
                 number: int | None = None):
        super().__init__()
        self.Label = labels or []
        self.Number = number

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(label={self.Label}, number={self.Number})'


class SystemSignalActivation(PropertyBasedPMType):
    """Represents BICEPS SystemSignalActivation."""

    NODETYPE = pm.SystemSignalActivation
    Manifestation: AlertSignalManifestation = cp.EnumAttributeProperty('Manifestation',
                                                                       enum_cls=AlertSignalManifestation,
                                                                       default_py_value=AlertSignalManifestation.OTH,
                                                                       is_optional=False)
    State: AlertActivation = cp.EnumAttributeProperty('State', default_py_value=AlertActivation.ON,
                                                      enum_cls=AlertActivation, is_optional=False)
    _props = ('Manifestation', 'State')

    def __init__(self, manifestation: AlertSignalManifestation | None = None,
                 state: AlertActivation | None = None):
        super().__init__()
        self.Manifestation = manifestation
        self.State = state

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(Manifestation={self.Manifestation}, State={self.State})'


class ProductionSpecification(PropertyBasedPMType):
    """Represents BICEPS AbstractDeviceComponentDescriptor/ProductionSpecification."""

    NODETYPE = pm.ProductionSpecification
    SpecType: CodedValue = cp.SubElementProperty(pm.SpecType, value_class=CodedValue)
    ProductionSpec: str = cp.NodeStringProperty(pm.ProductionSpec)
    ComponentId: InstanceIdentifier | None = cp.SubElementProperty(pm.ComponentId,
                                                                   value_class=InstanceIdentifier, is_optional=True)
    _props = ('SpecType', 'ProductionSpec', 'ComponentId')

    def __init__(self, spec_type: CodedValue | None = None,
                 production_spec: str | None = None,
                 component_id: InstanceIdentifier | None = None):
        super().__init__()
        self.SpecType = spec_type
        self.ProductionSpec = production_spec
        self.ComponentId = component_id


class BaseDemographics(PropertyBasedPMType):
    """Represents BICEPS BaseDemographics."""

    NODETYPE = pm.BaseDemographics
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Givenname: str | None = cp.NodeStringProperty(pm.Givenname, is_optional=True)
    Middlename: list[str] = cp.SubElementStringListProperty(pm.Middlename)
    Familyname: str | None = cp.NodeStringProperty(pm.Familyname, is_optional=True)
    Birthname: str | None = cp.NodeStringProperty(pm.Birthname, is_optional=True)
    Title: str | None = cp.NodeStringProperty(pm.Title, is_optional=True)
    _props = ('ExtExtension', 'Givenname', 'Middlename', 'Familyname', 'Birthname', 'Title')

    def __init__(self,
                 given_name: str | None = None,
                 middle_names: list[str] | None = None,
                 family_name: str | None = None,
                 birth_name: str | None = None,
                 title: str | None = None):
        super().__init__()
        self.Givenname = given_name
        self.Middlename = middle_names or []
        self.Familyname = family_name
        self.Birthname = birth_name
        self.Title = title


class PersonReference(PropertyBasedPMType):
    """Represents BICEPS PersonReference."""

    NODETYPE = pm.PersonReference
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Identification: list[InstanceIdentifier] = cp.SubElementListProperty(pm.Identification,
                                                                         value_class=InstanceIdentifier)  # 1...n
    Name: BaseDemographics | None = cp.SubElementProperty(pm.Name, value_class=BaseDemographics, is_optional=True)
    _props = ('ExtExtension', 'Identification', 'Name')

    def __init__(self, identifications: list[InstanceIdentifier] | None = None,
                 name: BaseDemographics | None = None):
        super().__init__()
        if identifications:
            self.Identification = identifications
        self.Name = name


class PersonParticipation(PersonReference):
    """Represents BICEPS PersonParticipation."""

    NODETYPE = pm.PersonParticipation
    Role: list[CodedValue] = cp.SubElementListProperty(pm.Role, value_class=CodedValue)
    _props = ('Role',)

    def __init__(self,
                 identifications: list[InstanceIdentifier] | None = None,
                 name: BaseDemographics | None = None,
                 roles: list[CodedValue] | None = None):
        super().__init__(identifications, name)
        if roles:
            self.Role = roles


class LocationDetail(PropertyBasedPMType):
    """Represents BICEPS LocationDetail."""

    NODETYPE = pm.LocationDetail
    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    PoC: str | None = cp.StringAttributeProperty('PoC')
    Room: str | None = cp.StringAttributeProperty('Room')
    Bed: str | None = cp.StringAttributeProperty('Bed')
    Facility: str | None = cp.StringAttributeProperty('Facility')
    Building: str | None = cp.StringAttributeProperty('Building')
    Floor: str | None = cp.StringAttributeProperty('Floor')
    _props = ('ExtExtension', 'PoC', 'Room', 'Bed', 'Facility', 'Building', 'Floor')

    def __init__(self,  # noqa: PLR0913
                 poc: str | None = None,
                 room: str | None = None,
                 bed: str | None = None,
                 facility: str | None = None,
                 building: str | None = None,
                 floor: str | None = None):
        super().__init__()
        self.PoC = poc
        self.Room = room
        self.Bed = bed
        self.Facility = facility
        self.Building = building
        self.Floor = floor


LocationDetailType = LocationDetail  # TypeAlias


class LocationReference(PropertyBasedPMType):
    """Represents BICEPS LocationReference."""

    NODETYPE = pm.LocationReference
    Identification: list[InstanceIdentifier] = cp.SubElementListProperty(pm.Identification,
                                                                         value_class=InstanceIdentifier)  # 1...n
    LocationDetail: LocationDetailType | None = cp.SubElementProperty(pm.LocationDetail,
                                                                      value_class=LocationDetail,
                                                                      is_optional=True)
    _props = ('Identification', 'LocationDetail')

    def __init__(self, identifications: list[InstanceIdentifier] | None = None,
                 location_detail: LocationDetailType | None = None):
        super().__init__()
        if identifications:
            self.Identification = identifications
        self.LocationDetail = location_detail


class ReferenceRange(PropertyBasedPMType):
    """Represents BICEPS ClinicalInfo/RelatedMeasurement/ReferenceRange."""

    NODETYPE = pm.ReferenceRange
    Range: RangeType = cp.SubElementProperty(pm.Range, value_class=Range)
    Meaning: CodedValue | None = cp.SubElementProperty(pm.Meaning, value_class=CodedValue, is_optional=True)
    _props = ('Range', 'Meaning')

    def __init__(self, ref_range: RangeType, meaning: CodedValue | None = None):
        super().__init__()
        self.Range = ref_range
        self.Meaning = meaning


ReferenceRangeType = ReferenceRange  # TypeAlias


class RelatedMeasurement(PropertyBasedPMType):
    """Represents BICEPS ClinicalInfo/RelatedMeasurement."""

    NODETYPE = pm.RelatedMeasurement
    Value: Measurement = cp.SubElementProperty(pm.Value, value_class=Measurement)
    ReferenceRange: list[ReferenceRangeType] = cp.SubElementListProperty(pm.ReferenceRange, value_class=ReferenceRange)
    _props = ('Value', 'ReferenceRange')

    def __init__(self,
                 value: Measurement,
                 reference_range: list[ReferenceRangeType] | None = None):
        super().__init__()
        self.Value = value
        if reference_range is not None:
            self.ReferenceRange = reference_range


class Criticality(StringEnum):
    """Represents BICEPS ClinicalInfo/Criticality."""

    Low = 'Lo'
    High = 'Hi'


CriticalityType = Criticality


class ClinicalInfo(PropertyBasedPMType):
    """Represents BICEPS ClinicalInfo."""

    NODETYPE = pm.ClinicalInfo
    Type: CodedValue | None = cp.SubElementProperty(pm.Type, value_class=CodedValue, is_optional=True)
    Code: CodedValue | None = cp.SubElementProperty(pm.Type, value_class=CodedValue, is_optional=True)
    Criticality: CriticalityType = cp.NodeEnumTextProperty(pm.Criticality, Criticality,
                                                           is_optional=True, implied_py_value=Criticality.Low)
    Description: list[LocalizedText] = cp.SubElementListProperty(pm.Description, value_class=LocalizedText)
    RelatedMeasurement: list[RelatedMeasurement] = cp.SubElementListProperty(pm.RelatedMeasurement,
                                                                             value_class=RelatedMeasurement)
    _props = ('Type', 'Code', 'Criticality', 'Description', 'RelatedMeasurement')

    def __init__(self,
                 type_: CodedValue | None = None,
                 code: CodedValue | None = None,
                 criticality: CriticalityType | None = None,
                 descriptions: list[LocalizedText] | None = None,
                 related_measurements: list[RelatedMeasurement] | None = None):
        super().__init__()
        self.Type = type_
        self.Code = code
        self.Criticality = criticality
        if descriptions:
            self.Description = descriptions
        if related_measurements:
            self.RelatedMeasurement = related_measurements


ClinicalInfoType = ClinicalInfo  # TypeAlias


class ImagingProcedure(PropertyBasedPMType):
    """Represents BICEPS ImagingProcedure."""

    NODETYPE = pm.ImagingProcedure
    AccessionIdentifier: InstanceIdentifier = cp.SubElementProperty(pm.AccessionIdentifier,
                                                                    value_class=InstanceIdentifier)
    RequestedProcedureId: InstanceIdentifier = cp.SubElementProperty(pm.RequestedProcedureId,
                                                                     value_class=InstanceIdentifier)
    StudyInstanceUid: InstanceIdentifier = cp.SubElementProperty(pm.StudyInstanceUid, value_class=InstanceIdentifier)
    ScheduledProcedureStepId: InstanceIdentifier = cp.SubElementProperty(pm.ScheduledProcedureStepId,
                                                                         value_class=InstanceIdentifier)
    Modality: CodedValue | None = cp.SubElementProperty(pm.Modality, value_class=CodedValue, is_optional=True)
    ProtocolCode: CodedValue | None = cp.SubElementProperty(pm.ProtocolCode, value_class=CodedValue, is_optional=True)
    _props = ('AccessionIdentifier', 'RequestedProcedureId', 'StudyInstanceUid', 'ScheduledProcedureStepId',
              'Modality', 'ProtocolCode')

    def __init__(self,  # noqa: PLR0913
                 accession_identifier: InstanceIdentifier | None = None,
                 requested_procedure_id: InstanceIdentifier | None = None,
                 study_instance_uid: InstanceIdentifier | None = None,
                 scheduled_procedure_step_id: InstanceIdentifier | None = None,
                 modality: InstanceIdentifier | None = None,
                 protocol_code: CodedValue | None = None):
        super().__init__()
        self.AccessionIdentifier = accession_identifier
        self.RequestedProcedureId = requested_procedure_id
        self.StudyInstanceUid = study_instance_uid
        self.ScheduledProcedureStepId = scheduled_procedure_step_id
        self.Modality = modality
        self.ProtocolCode = protocol_code


ImagingProcedureType = ImagingProcedure  # TypeAlias


class OrderDetail(PropertyBasedPMType):
    """Represents BICEPS OrderDetail."""

    NODETYPE = pm.OrderDetail
    Start: str | None = cp.NodeStringProperty(pm.Start, is_optional=True)  # xsd:dateTime
    End: str | None = cp.NodeStringProperty(pm.End, is_optional=True)  # xsd:dateTime
    Performer: list[PersonParticipation] = cp.SubElementListProperty(pm.Performer, value_class=PersonParticipation)
    Service: list[CodedValue] = cp.SubElementListProperty(pm.Service, value_class=CodedValue)
    ImagingProcedure: list[ImagingProcedureType] = cp.SubElementListProperty(pm.ImagingProcedure,
                                                                             value_class=ImagingProcedure)
    _props = ('Start', 'End', 'Performer', 'Service', 'ImagingProcedure')

    def __init__(self,
                 start: str | None = None,
                 end: str | None = None,
                 performer: list[PersonParticipation] | None = None,
                 service: list[CodedValue] | None = None,
                 imaging_procedure: list[ImagingProcedureType] | None = None):
        super().__init__()
        self.Start = start
        self.End = end
        if performer:
            self.Performer = performer
        if service:
            self.Service = service
        if imaging_procedure:
            self.ImagingProcedure = imaging_procedure


class RequestedOrderDetail(OrderDetail):
    """Represents BICEPS WorkflowContextState/WorkflowDetail/RequestedOrderDetail."""

    NODETYPE = pm.RequestedOrderDetail
    ReferringPhysician: PersonReference | None = cp.SubElementProperty(pm.ReferringPhysician,
                                                                       value_class=PersonReference,
                                                                       is_optional=True)
    RequestingPhysician: PersonReference | None = cp.SubElementProperty(pm.RequestingPhysician,
                                                                        value_class=PersonReference,
                                                                        is_optional=True)
    PlacerOrderNumber: InstanceIdentifier = cp.SubElementProperty(pm.PlacerOrderNumber, value_class=InstanceIdentifier)
    _props = ('ReferringPhysician', 'RequestingPhysician', 'PlacerOrderNumber')

    def __init__(self,  # noqa: PLR0913
                 start: str | None = None,
                 end: str | None = None,
                 performer: list[PersonParticipation] | None = None,
                 service: list[CodedValue] | None = None,
                 imaging_procedure: list[ImagingProcedure] | None = None,
                 referring_physician: PersonReference | None = None,
                 requesting_physician: PersonReference | None = None,
                 placer_order_number: InstanceIdentifier | None = None):
        super().__init__(start, end, performer, service, imaging_procedure)
        self.ReferringPhysician = referring_physician
        self.RequestingPhysician = requesting_physician
        self.PlacerOrderNumber = placer_order_number


RequestedOrderDetailType = RequestedOrderDetail  # TypeAlias


class PerformedOrderDetail(OrderDetail):
    """Represents BICEPS WorkflowContextState/WorkflowDetail/PerformedOrderDetail."""

    NODETYPE = pm.PerformedOrderDetail
    FillerOrderNumber: InstanceIdentifier | None = cp.SubElementProperty(pm.FillerOrderNumber,
                                                                         value_class=InstanceIdentifier,
                                                                         is_optional=True)
    ResultingClinicalInfo: ClinicalInfoType = cp.SubElementListProperty(pm.RelevantClinicalInfo,
                                                                        value_class=ClinicalInfo)
    _props = ('FillerOrderNumber', 'ResultingClinicalInfo')

    def __init__(self,  # noqa: PLR0913
                 start: str | None = None,
                 end: str | None = None,
                 performer: list[PersonParticipation] | None = None,
                 service: list[CodedValue] | None = None,
                 imaging_procedure: list[ImagingProcedure] | None = None,
                 filler_order_number: InstanceIdentifier | None = None,
                 resulting_clinical_info: ClinicalInfoType | None = None):
        super().__init__(start, end, performer, service, imaging_procedure)
        self.FillerOrderNumber = filler_order_number
        if resulting_clinical_info:
            self.ResultingClinicalInfo = resulting_clinical_info


PerformedOrderDetailType = PerformedOrderDetail  # TypeAlias


class WorkflowDetail(PropertyBasedPMType):
    """Represents BICEPS WorkflowContextState/WorkflowDetail."""

    NODETYPE = pm.WorkflowDetail
    Patient = cp.SubElementProperty(pm.Patient, value_class=PersonReference)
    AssignedLocation: LocationReference | None = cp.SubElementProperty(pm.AssignedLocation,
                                                                       value_class=LocationReference,
                                                                       is_optional=True)
    VisitNumber: InstanceIdentifier | None = cp.SubElementProperty(pm.VisitNumber,
                                                                   value_class=InstanceIdentifier, is_optional=True)
    DangerCode: list[CodedValue] = cp.SubElementListProperty(pm.DangerCode, value_class=CodedValue)
    RelevantClinicalInfo: list[ClinicalInfoType] = cp.SubElementListProperty(pm.RelevantClinicalInfo,
                                                                             value_class=ClinicalInfo)
    RequestedOrderDetail: RequestedOrderDetailType | None = cp.SubElementProperty(pm.RequestedOrderDetail,
                                                                                  value_class=RequestedOrderDetail,
                                                                                  is_optional=True)
    PerformedOrderDetail: PerformedOrderDetailType = cp.SubElementProperty(pm.PerformedOrderDetail,
                                                                           value_class=PerformedOrderDetail,
                                                                           is_optional=True)
    _props = ('Patient', 'AssignedLocation', 'VisitNumber', 'DangerCode',
              'RelevantClinicalInfo', 'RequestedOrderDetail', 'PerformedOrderDetail')

    def __init__(self,  # noqa: PLR0913
                 patient: PersonReference | None = None,
                 assigned_location: LocationReference | None = None,
                 visit_number: InstanceIdentifier | None = None,
                 danger_code: list[CodedValue] | None = None,
                 relevant_clinical_info: list[ClinicalInfo] | None = None,
                 requested_order_detail: RequestedOrderDetailType | None = None,
                 performed_order_detail: PerformedOrderDetailType | None = None):
        super().__init__()
        self.Patient = patient
        self.AssignedLocation = assigned_location
        self.VisitNumber = visit_number
        if danger_code:
            self.DangerCode = danger_code
        if relevant_clinical_info:
            self.RelevantClinicalInfo = relevant_clinical_info
        self.RequestedOrderDetail = requested_order_detail
        self.PerformedOrderDetail = performed_order_detail


class Relation(PropertyBasedPMType):
    """Represents BICEPS AbstractMetricDescriptor/Relation.

    Relation allows the modelling of relationships between a metric and other containment tree entries.
    """

    ExtExtension = cp.ExtensionNodeProperty(ext.Extension)
    Code: CodedValue | None = cp.SubElementProperty(pm.Code, value_class=CodedValue, is_optional=True)
    Identification: InstanceIdentifier | None = cp.SubElementProperty(pm.Identification,
                                                                      value_class=InstanceIdentifier,
                                                                      is_optional=True)
    Kind: AbstractMetricDescriptorRelationKindEnum | None = cp.EnumAttributeProperty(
        'Kind',
        enum_cls=AbstractMetricDescriptorRelationKindEnum,
        is_optional=False)
    Entries: list[str] = cp.EntryRefListAttributeProperty('Entries')
    _props = ('ExtExtension', 'Code', 'Identification', 'Kind', 'Entries')


class PatientDemographicsCoreData(BaseDemographics):
    """Represents BICEPS PatientDemographicsCoreData."""

    NODETYPE = pm.PatientDemographicsCoreData
    Sex: SexType | None = cp.NodeEnumTextProperty(pm.Sex, Sex, is_optional=True)
    PatientType: PatientTypeType | None = cp.NodeEnumTextProperty(pm.PatientType, PatientType, is_optional=True)
    DateOfBirth: DateTypeUnion | None = cp.DateOfBirthProperty(pm.DateOfBirth, is_optional=True)
    Height: Measurement | None = cp.SubElementProperty(pm.Height, value_class=Measurement, is_optional=True)
    Weight: Measurement | None = cp.SubElementProperty(pm.Weight, value_class=Measurement, is_optional=True)
    Race: CodedValue | None = cp.SubElementProperty(pm.Race, value_class=CodedValue, is_optional=True)
    _props = ('Sex', 'PatientType', 'DateOfBirth', 'Height', 'Weight', 'Race')

    def set_birthdate(self, date_time_of_birth_string: str):
        """Format string acc. to XML Schema: xsd:dateTime, xsd:date, xsd:gYearMonth or xsd:gYear.

        Internally it holds it as a datetime object, so specific formatting of the dateTimeOfBirth_string will be lost.
        """
        if not date_time_of_birth_string:
            self.DateOfBirth = None
        else:
            self.DateOfBirth = cp.DateOfBirthProperty.mk_value_object(date_time_of_birth_string)


class NeonatalPatientDemographicsCoreData(PatientDemographicsCoreData):
    """Represents BICEPS NeonatalPatientDemographicsCoreData."""

    NODETYPE = pm.NeonatalPatientDemographicsCoreData
    GestationalAge: Measurement | None = cp.SubElementProperty(pm.GestationalAge,
                                                               value_class=Measurement,
                                                               is_optional=True)
    BirthLength: Measurement | None = cp.SubElementProperty(pm.BirthLength, value_class=Measurement, is_optional=True)
    BirthWeight: Measurement | None = cp.SubElementProperty(pm.BirthWeight, value_class=Measurement, is_optional=True)
    HeadCircumference: Measurement | None = cp.SubElementProperty(pm.HeadCircumference,
                                                                  value_class=Measurement,
                                                                  is_optional=True)
    Mother: PersonReference | None = cp.SubElementProperty(pm.Mother, value_class=PersonReference, is_optional=True)
    _props = ('GestationalAge', 'BirthLength', 'BirthWeight', 'HeadCircumference', 'Mother')


class Udi(PropertyBasedPMType):
    """Represents BICEPS MdsDescriptor/MetaData/Udi."""

    NODETYPE = pm.Udi
    DeviceIdentifier: str = cp.NodeStringProperty(pm.DeviceIdentifier)
    HumanReadableForm: str = cp.NodeStringProperty(pm.HumanReadableForm)
    Issuer: InstanceIdentifier = cp.SubElementProperty(pm.Issuer, value_class=InstanceIdentifier)
    Jurisdiction: InstanceIdentifier | None = cp.SubElementProperty(pm.Jurisdiction,
                                                                    value_class=InstanceIdentifier,
                                                                    is_optional=True)
    _props = ('DeviceIdentifier', 'HumanReadableForm', 'Issuer', 'Jurisdiction')

    def __init__(self, device_identifier: str | None = None,
                 human_readable_form: str | None = None,
                 issuer: InstanceIdentifier | None = None,
                 jurisdiction: InstanceIdentifier | None = None):
        super().__init__()
        self.DeviceIdentifier = device_identifier
        self.HumanReadableForm = human_readable_form
        self.Issuer = issuer
        self.Jurisdiction = jurisdiction


UdiType = Udi  # TypeAlias


class MetaData(PropertyBasedPMType):
    """Represents BICEPS MdsDescriptor/MetaData."""

    NODETYPE = pm.MetaData
    Udi: list[UdiType] = cp.SubElementListProperty(pm.Udi, value_class=UdiType)
    LotNumber: str | None = cp.NodeStringProperty(pm.LotNumber, is_optional=True)
    Manufacturer: list[LocalizedText] = cp.SubElementListProperty(pm.Manufacturer, value_class=LocalizedText)
    ManufactureDate: str | None = cp.NodeStringProperty(pm.ManufactureDate, is_optional=True)
    ExpirationDate: str | None = cp.NodeStringProperty(pm.ExpirationDate, is_optional=True)
    ModelName: list[LocalizedText] = cp.SubElementListProperty(pm.ModelName, value_class=LocalizedText)
    ModelNumber: str | None = cp.NodeStringProperty(pm.ModelNumber, is_optional=True)
    SerialNumber: list[str] = cp.SubElementStringListProperty(pm.SerialNumber)
    _props = ('Udi', 'LotNumber', 'Manufacturer', 'ManufactureDate', 'ExpirationDate',
              'ModelName', 'ModelNumber', 'SerialNumber')


class CalibrationResult(PropertyBasedPMType):
    """Represents BICEPS CalibrationInfo/CalibrationDocumentation/CalibrationResult."""

    NODETYPE = pm.CalibrationResult
    Code: CodedValue = cp.SubElementProperty(pm.Code, value_class=CodedValue)
    Value: Measurement = cp.SubElementProperty(pm.Value, value_class=Measurement)
    _props = ('Code', 'Value')


class CalibrationDocumentation(PropertyBasedPMType):
    """Represents BICEPS CalibrationInfo/CalibrationDocumentation."""

    NODETYPE = pm.CalibrationDocumentation
    Documentation: list[LocalizedText] = cp.SubElementListProperty(pm.Documentation, value_class=LocalizedText)
    CalibrationResult: list[CalibrationResult] = cp.SubElementListProperty(pm.CalibrationResult,
                                                                           value_class=CalibrationResult)
    _props = ('Documentation', 'CalibrationResult')


class CalibrationInfo(PropertyBasedPMType):
    """Represents BICEPS CalibrationInfo."""

    NODETYPE = pm.CalibrationInfo
    CalibrationDocumentation: list[CalibrationDocumentation] = cp.SubElementListProperty(
        pm.CalibrationDocumentation, value_class=CalibrationDocumentation)
    ComponentCalibrationState: CalibrationState | None = cp.EnumAttributeProperty('ComponentCalibrationState',
                                                                                  enum_cls=CalibrationState)
    Type: CalibrationType | None = cp.EnumAttributeProperty('Type',
                                                            implied_py_value=CalibrationType.UNSPEC,
                                                            enum_cls=CalibrationType)
    Time: float | None = cp.TimestampAttributeProperty('Time')
    _props = ('CalibrationDocumentation', 'ComponentCalibrationState', 'Type', 'Time')


class ApprovedJurisdictions(PropertyBasedPMType):
    """Represents BICEPS ApprovedJurisdictions."""

    NODETYPE = pm.ApprovedJurisdictions
    ApprovedJurisdiction: list[InstanceIdentifier] = cp.SubElementListProperty(pm.ApprovedJurisdiction,
                                                                               value_class=InstanceIdentifier)
    _props = ('ApprovedJurisdiction',)


class ContainmentTreeEntry(PropertyBasedPMType):
    """Represents BICEPS ContainmentTree/Entry."""

    Type: CodedValue | None = cp.SubElementProperty(pm.Type, value_class=CodedValue, is_optional=True)
    _props = ('Type',)


class ContainmentTree(PropertyBasedPMType):
    """Represents BICEPS ContainmentTree."""

    Entry: ContainmentTreeEntry = cp.SubElementListProperty(pm.Entry, value_class=ContainmentTreeEntry)
    _props = ('Entry',)


# Technically, Retrievalibity belongs to msg_types, because it is defined in BICEPS message model.
# But its usage is in mdib, not in messages, it fits here better.
# This also helps to avoid a circular dependency.
class RetrievabilityMethod(StringEnum):
    """Represents BICEPS message model RetrievabilityMethod."""

    GET = 'Get'
    PERIODIC = 'Per'
    EPISODIC = 'Ep'
    STREAM = 'Strm'


class RetrievabilityInfo(PropertyBasedPMType):
    """Represents BICEPS message model RetrievabilityInfo."""

    NODETYPE = msg.RetrievabilityInfo
    Method: RetrievabilityMethod = cp.EnumAttributeProperty('Method',
                                                            enum_cls=RetrievabilityMethod,
                                                            is_optional=False)
    UpdatePeriod: DurationType | None = cp.DurationAttributeProperty('UpdatePeriod', implied_py_value=1.0)
    _props = ('Method', 'UpdatePeriod')

    def __init__(self, method: RetrievabilityMethod, update_period: DurationType | None = None):
        super().__init__()
        self.Method = method
        self.UpdatePeriod = update_period

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> RetrievabilityInfo:
        """Construct class from a node."""
        obj = cls(RetrievabilityMethod.GET)  # any allowed value, will be overwritten in update_node
        obj.update_from_node(node)
        return obj

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} {self.Method} period={self.UpdatePeriod}'


class Retrievability(PropertyBasedPMType):
    """Represents BICEPS message model Retrievability."""

    By: list[RetrievabilityInfo] = cp.SubElementListProperty(msg.By, value_class=RetrievabilityInfo)
    _props = ('By',)

    def __init__(self, retrievability_info_list: list[RetrievabilityInfo] | None = None):
        super().__init__()
        if retrievability_info_list:
            self.By = retrievability_info_list


# mapping of types: xsi:type information to classes
# find all classes in this module that have a member "NODETYPE"
classes = inspect.getmembers(sys.modules[__name__],
                             lambda member: inspect.isclass(member) and member.__module__ == __name__)

# make a dictionary from found classes: (Key is NODETYPE, value is the class itself
_classes = [c[1] for c in classes if hasattr(c[1], 'NODETYPE') and c[1].NODETYPE is not None]

_name_class_lookup = {c.NODETYPE: c for c in _classes}


def _get_pmtypes_class(qname: etree.QName) -> type[PropertyBasedPMType]:
    """Find class in _name_class_lookup."""
    try:
        return _name_class_lookup[qname]
    except KeyError as ex:
        raise KeyError(f'{qname.namespace}.{qname.localname}') from ex  # noqa: EM102



from __future__ import annotations

import logging
import socket
import threading
import traceback
from dataclasses import dataclass
from http.server import HTTPServer

from .httprequesthandler import DispatchingRequestHandler
from sdc11073.dispatch import PathElementRegistry
from sdc11073.loghelper import LoggerAdapter


@dataclass(frozen=True)
class _ThreadInfo:
    thread: threading.Thread
    request: socket.socket
    client_address: tuple


class _ThreadingHTTPServer(HTTPServer):
    """ Each request is handled in a thread.
    """

    def __init__(self, logger, server_address,
                 chunk_size, supported_encodings):
        self.daemon_threads = True
        self.threads = []
        self.logger = logger
        self.dispatcher = PathElementRegistry()
        self.chunk_size = chunk_size
        self.supported_encodings = supported_encodings
        super().__init__(server_address, DispatchingRequestHandler)

    def process_request_thread(self, request, client_address):
        """Same as in BaseServer but as a thread.
        """
        try:
            self.finish_request(request, client_address)
        except (ConnectionResetError, ConnectionAbortedError) as ex:
            self.logger.info('Connection reset by {}: {}', client_address, ex)
        except Exception:
            self.handle_error(request, client_address)
        finally:
            self.shutdown_request(request)
            # this thread will close after return from this method, it can already be removed from self.threads
            for thread_info in self.threads:
                if thread_info.request == request:
                    self.threads.remove(thread_info)

    def process_request(self, request, client_address):
        """Start a new thread to process the request."""
        thread = threading.Thread(target=self.process_request_thread,
                                  args=(request, client_address),
                                  name=f'SubscrRecv{client_address}')
        thread.daemon = True
        self.threads.append(_ThreadInfo(thread, request, client_address))
        thread.start()

    def server_close(self):
        super().server_close()
        if self.dispatcher is not None:
            self.dispatcher.methods = {}
            self.dispatcher = None  # this leads to a '503' reaction in SOAPNotificationsHandler
        for thread_info in self.threads:
            if thread_info.thread.is_alive():
                try:
                    thread_info.request.shutdown(socket.SHUT_RDWR)
                    thread_info.request.close()
                    self.logger.info('closed socket for notifications from {}', thread_info.client_address)
                except OSError:
                    # the connection is already closed
                    continue
                except Exception as ex:
                    self.logger.warn('error closing socket for notifications from {}: {}', thread_info.client_address,
                                     ex)


class HttpServerThreadBase(threading.Thread):

    def __init__(self, my_ipaddress, ssl_context, supported_encodings,
                 logger, chunk_size=0):
        """
        Runs a ThreadingHTTPServer in a thread, so that it can be stopped without blocking.
        Handling of requests happens in two stages:
        - the http server instantiates a request handler with the request
        - the request handler forwards the handling itself to a dispatcher (due to the dynamic nature of the handling).
        :param my_ipaddress: The ip address that the http server shall bind to (no port!)
        :param ssl_context: a ssl.SslContext instance or None
        :param supported_encodings: a list of strings
        :param logger: a python logger
        :param chunk_size: if value > 0, messages are split into chunks of this size.
        """
        super().__init__(name='Dev_SdcHttpServerThread')
        self.daemon = True

        self._my_ipaddress = my_ipaddress
        self._ssl_context = ssl_context
        self.my_port = None
        self.httpd = None
        self.supported_encodings = supported_encodings
        if isinstance(logger, logging.Logger):
            self.logger = LoggerAdapter(logger)
        else:
            self.logger = logger
        self.chunk_size = chunk_size
        # create and set up the dispatcher for all incoming requests
        self.started_evt = threading.Event()  # helps to wait until thread has initialised is variables
        self._stop_requested = False
        self.base_url = None

    def run(self):
        self._stop_requested = False
        try:
            myport = 0  # zero means that OS selects a free port
            self.httpd = _ThreadingHTTPServer(self.logger,
                                              (self._my_ipaddress, myport),
                                              self.chunk_size,
                                              self.supported_encodings)
            self.my_port = self.httpd.server_port
            self.logger.info('starting http server on {}:{}', self._my_ipaddress, self.my_port)
            if self._ssl_context:
                self.httpd.socket = self._ssl_context.wrap_socket(self.httpd.socket, server_side=True)
                self.base_url = f'https://{self._my_ipaddress}:{self.my_port}/'
            else:
                self.base_url = f'http://{self._my_ipaddress}:{self.my_port}/'

            self.started_evt.set()
            self.httpd.serve_forever()
        except Exception:
            if not self._stop_requested:
                self.logger.error(
                    f'Unhandled Exception at thread runtime. Thread will abort! {traceback.format_exc()}')
            raise
        finally:
            self.logger.info('http server stopped.')

    @property
    def dispatcher(self):
        return self.httpd.dispatcher

    def stop(self):
        self._stop_requested = True
        self.httpd.shutdown()
        self.httpd.server_close()
        for thread_info in self.httpd.threads:
            if thread_info.thread.is_alive():
                thread_info.thread.join(1)
            if thread_info.thread.is_alive():
                self.logger.warn('could not end client thread for notifications from {}', thread_info.client_address)
        del self.httpd.threads[:]



from __future__ import annotations

import logging
import traceback
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from .request import RequestData
from sdc11073 import commlog
from sdc11073.exceptions import HTTPRequestHandlingError
from sdc11073.pysoap.soapenvelope import Fault, faultcodeEnum
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

if TYPE_CHECKING:
    from .dispatchkey import RequestHandlerProtocol


class MessageConverterMiddleware:
    """ Converts between http server message format and internal format
    http server is strings, internal is RequestData."""

    def __init__(self, msg_reader, msg_factory, logger, dispatcher: RequestHandlerProtocol):
        self._logger = logger
        self._msg_reader = msg_reader
        self._msg_factory = msg_factory
        self._dispatcher: RequestHandlerProtocol = dispatcher
        self._soap_request_in_logger = logging.getLogger(commlog.SOAP_REQUEST_IN)
        self._soap_response_out_logger = logging.getLogger(commlog.SOAP_RESPONSE_OUT)

    def do_post(self, headers: dict, path: str, peer_name: str, request_bytes: bytes) -> (int, str, str):
        http_status = 200
        http_reason = 'Ok'
        response_xml_string = 'not set yet'
        self._soap_request_in_logger.debug(request_bytes, extra={'https_method': 'POST'})

        # try to read the request
        fault = None
        message_data = None
        try:
            message_data = self._msg_reader.read_received_message(request_bytes)
        except HTTPRequestHandlingError as ex:
            self._logger.warning('could not read message: {}', str(ex))
            fault = ex.soap_fault
            http_status = ex.status
            http_reason = ex.reason
        except Exception as ex:
            http_status = 500
            http_reason = 'exception'
            self._logger.warning(traceback.format_exc())
            fault = Fault()
            fault.Code.Value = faultcodeEnum.SENDER
            fault.add_reason_text(str(ex))

        if fault is not None:
            inf = HeaderInformationBlock(action=fault.action, addr_to=None)
            response = self._msg_factory.mk_soap_message(inf, payload=fault)
            response_xml_string = response.serialize()
            self._soap_response_out_logger.debug(response_xml_string, extra={'http_method': 'POST'})
            return http_status, http_reason, response_xml_string

        # handle the request
        try:
            request_data = RequestData(headers, path, peer_name, request_bytes, message_data)
            request_data.consume_current_path_element()  # uuid is already used
            response = self._dispatcher.on_post(request_data)
            response_xml_string = response.serialize()
        except HTTPRequestHandlingError as ex:
            message_data = self._msg_reader.read_received_message(request_bytes, validate=False)
            request_data = RequestData(headers, path, peer_name, request_bytes, message_data)
            response = self._msg_factory.mk_reply_soap_message(request_data, ex.soap_fault)
            response_xml_string = response.serialize()
            http_status = ex.status
            http_reason = ex.reason
        except Exception as ex:
            # make an error 500 response with the soap fault as content
            self._logger.error(traceback.format_exc())
            message_data = self._msg_reader.read_received_message(request_bytes, validate=False)
            request_data = RequestData(headers, path, peer_name, request_bytes, message_data)
            fault = Fault()
            fault.Code.Value = faultcodeEnum.SENDER
            fault.add_reason_text(str(ex))
            response = self._msg_factory.mk_reply_soap_message(request_data, fault)
            response_xml_string = response.serialize()
            http_status = 500
            http_reason = 'exception'
        finally:
            self._soap_response_out_logger.debug(response_xml_string, extra={'http_method': 'POST'})
            return http_status, http_reason, response_xml_string  # noqa: B012

    def do_get(self, headers: dict, path: str, peer_name: str) -> (int, str, str, str):
        parsed_path = urlparse(path)
        try:
            # GET has no content, log it to document duration of processing
            self._soap_request_in_logger.debug(b'', extra={'http_method': 'GET'})
            request_data = RequestData(headers, path, peer_name)
            request_data.consume_current_path_element()  # uuid is already used
            response_string = self._dispatcher.on_get(request_data)
            self._soap_response_out_logger.debug(response_string, extra={'http_method': 'GET'})
            if parsed_path.query == 'wsdl':
                content_type = "text/xml; charset=utf-8"
            else:
                content_type = "application/soap+xml; charset=utf-8"
            return 200, 'Ok', response_string, content_type
        except Exception as ex:
            self._logger.error(traceback.format_exc())
            response_string = str(ex).encode('utf-8')
            content_type = "text"
            return 500, 'Exception', response_string, content_type
